import argparse
import datetime
import json
import logging
import sys
from pathlib import Path
from typing import Any, Dict, List, Optional

import annofabapi
import dateutil.parser
import pandas as pd

import annofabcli
import annofabcli.common.cli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import (
    AbstractCommandLineInterface,
    ArgumentParser,
    build_annofabapi_resource_and_login,
    get_list_from_args,
)
from annofabcli.common.download import DownloadingFile
from annofabcli.common.utils import read_lines_except_blank_line

logger = logging.getLogger(__name__)


def get_err_history_events(
    task_history_events: List[Dict[str, Any]], task_id_list: List[str]
) -> Dict[str, List[Dict[str, Any]]]:
    """
    対象のtask_idが含まれるhistory_eventsを返す
    """
    err_history_events_dict: Dict[str, List[Dict[str, Any]]] = {}

    for task_history_event in task_history_events:
        if task_id_list:
            if task_history_event["task_id"] in task_id_list:
                if task_history_event["task_id"] in err_history_events_dict:
                    err_history_events_dict[task_history_event["task_id"]].append(task_history_event)
                else:
                    err_history_events_dict[task_history_event["task_id"]] = [task_history_event]
        else:
            if task_history_event["task_id"] in err_history_events_dict:
                err_history_events_dict[task_history_event["task_id"]].append(task_history_event)
            else:
                err_history_events_dict[task_history_event["task_id"]] = [task_history_event]

    return err_history_events_dict


class FindBreakError(AbstractCommandLineInterface):
    def __init__(self, service: annofabapi.Resource, facade: AnnofabApiFacade, args: argparse.Namespace):
        super().__init__(service, facade, args)
        project_id_list = get_list_from_args(args.project_id)
        self.project_id_list = list(set(project_id_list))

    def _get_username(self, project_id: str, account_id: str) -> Optional[str]:
        """
        プロジェクトメンバのusernameを取得する。プロジェクトメンバでなければ、account_idを返す。
        account_idがNoneならばNoneを返す。
        """
        member = self.facade.get_project_member_from_account_id(project_id, account_id)
        if member is not None:
            return member["username"]
        else:
            return account_id

    def _get_all_tasks(self, project_id: str, task_query: Optional[Dict[str, Any]] = None) -> List[Dict[str, Any]]:
        """
        task一覧を取得する
        """
        tasks = self.service.wrapper.get_all_tasks(project_id, query_params=task_query)
        return tasks

    def _project_task_history_events(
        self, project_id: str, import_file_path: Optional[str] = None
    ) -> Optional[List[Dict[str, Any]]]:
        """
        タスク履歴イベント全件ファイルを取得する。
        import_fileがNone:history_events_urlパスから直接読み込む
        import_fileがNoneではない:import_fileで指定されたファイルから読み込む
        """
        project_task_history_events: List[Dict[str, Any]] = []
        if import_file_path is None:
            downloading_obj = DownloadingFile(self.service)
            cache_dir = annofabcli.utils.get_cache_dir()
            json_path = cache_dir / f"{project_id}-inspection.json"
            downloading_obj.download_task_history_event_json(project_id, str(json_path))
            with json_path.open() as f:
                project_task_history_events = json.load(f)

        else:
            history_events = read_lines_except_blank_line(import_file_path)
            project_task_history_events = json.loads(history_events[0])

        return project_task_history_events

    def get_err_events(
        self, err_history_events: Dict[str, List[Dict[str, Any]]], time_list: List[str]
    ) -> List[List[Dict[str, Any]]]:
        """
        しきい値以上の作業時間になっている開始と終了のhistory_eventsのペアを返す
        """

        def check_applicable_time(
            from_time: datetime.datetime, to_time: datetime.datetime, date_time_list: List[datetime.datetime]
        ):
            # timeが検索対象と合致しているかを探す
            from_time = dateutil.parser.parse(from_time.strftime("%Y-%m-%d %H:%M:%S"))
            to_time = dateutil.parser.parse(to_time.strftime("%Y-%m-%d %H:%M:%S"))

            if date_time_list:
                for date_time in date_time_list:
                    date_time = dateutil.parser.parse(date_time.strftime("%Y-%m-%d %H:%M:%S"))
                    if date_time in [from_time, to_time]:
                        return True
            else:
                return True
            return False

        err_events_list = []
        date_time_list = [dateutil.parser.parse(time) for time in time_list]
        for v in err_history_events.values():
            v.sort(key=lambda x: x["created_datetime"])
            for i, history_events in enumerate(v):
                if history_events["status"] == "working":
                    if len(v) == i + 1:
                        continue
                    next_history_events = v[i + 1]
                    if next_history_events["status"] in ["on_hold", "break", "complete"]:
                        next_time = dateutil.parser.parse(next_history_events["created_datetime"])
                        this_time = dateutil.parser.parse(history_events["created_datetime"])
                        working_time = next_time - this_time
                        if working_time > datetime.timedelta(
                            minutes=self.args.task_history_time_threshold
                        ) and check_applicable_time(this_time, next_time, date_time_list):
                            history_events["user_name"] = self._get_username(
                                history_events["project_id"], history_events["account_id"]
                            )
                            next_history_events["user_name"] = self._get_username(
                                history_events["project_id"], history_events["account_id"]
                            )
                            history_events["project_title"] = self.facade.get_project_title(
                                history_events["project_id"]
                            )
                            next_history_events["project_title"] = self.facade.get_project_title(
                                history_events["project_id"]
                            )
                            err_events_list.append([history_events, next_history_events])
        return err_events_list

    @staticmethod
    def validate(args: argparse.Namespace) -> bool:
        COMMON_MESSAGE = "annofabcli experimental find_break_error: error:"
        if args.import_file_path is not None:
            if not Path(args.import_file_path).is_file():
                print(
                    f"{COMMON_MESSAGE} argument --import_file_path: ファイルパスが存在しません\
                '{args.import_file_path}'",
                    file=sys.stderr,
                )
                return False
        project_id_list = get_list_from_args(args.project_id)
        if len(project_id_list) > 1 and args.task_id:
            print(
                f"{COMMON_MESSAGE} argument --project_id: task_idを指定した場合はproject_idは複数指定できません\
                            '{args.project_id}'",
                file=sys.stderr,
            )
            return False
        if args.task_id and len(args.task_id) > 1 and args.time:
            print(
                f"{COMMON_MESSAGE} argument --task_id: timeを指定した場合はtask_idは複数指定できません\
                            '{args.project_id}'",
                file=sys.stderr,
            )
            return False

        return True

    def main(self):
        args = self.args
        if not self.validate(args):
            return

        err_events = []
        for project_id in args.project_id:
            task_history_events = self._project_task_history_events(
                project_id=project_id, import_file_path=args.import_file_path
            )
            if not task_history_events:
                # 全件ファイルの更新に失敗したらスルー
                continue
            # task_id 絞り込み
            err_history_events = get_err_history_events(
                task_history_events=task_history_events, task_id_list=args.task_id if args.task_id else []
            )
            # timeとしきい値絞り込み
            err_events.extend(
                self.get_err_events(err_history_events=err_history_events, time_list=args.time if args.time else [])
            )

        output_err_events(err_events_list=err_events, output=self.output, add=args.add)


def output_err_events(err_events_list: List[List[Dict[str, Any]]], output: str = None, add: bool = False):
    """
    開始と終了のhistory_eventsのペアから出力する
    :param err_events_list:
    :param output:
    :return:
    """

    def _timedelta_to_HM(td: datetime.timedelta):
        sec = td.total_seconds()
        return str(round(sec // 3600)) + "時間" + str(round(sec % 3600 // 60)) + "分"

    data_list = []
    for i, data in enumerate(err_events_list):
        start_data, end_data = data
        start_time = dateutil.parser.parse(start_data["created_datetime"])
        end_time = dateutil.parser.parse(end_data["created_datetime"])
        start_data["datetime"] = start_time.isoformat()
        end_data["datetime"] = end_time.isoformat()
        start_data["working_time"] = ""
        end_data["working_time"] = _timedelta_to_HM(end_time - start_time)

        df = pd.DataFrame([start_data, end_data])
        df["no"] = i + 1

        del df["phase_stage"]
        del df["account_id"]
        del df["created_datetime"]
        data_list.append(df)

    if not data_list:
        logger.warning("historyが見つからなかった為、出力しません。")
        return
    elif len(data_list) == 1:
        pd_data = data_list[0]
    else:
        pd_data = pd.concat(data_list)

    if add:
        to_csv_kwargs = {
            "index": False,
            "mode": "a",
            "header": False,
            "encoding": "utf_8_sig",
            "line_terminator": "\r\n",
        }
    else:
        to_csv_kwargs = {"index": False, "encoding": "utf_8_sig", "line_terminator": "\r\n"}

    annofabcli.utils.print_csv(
        pd_data[
            [
                "no",
                "project_title",
                "project_id",
                "task_id",
                "user_name",
                "phase",
                "status",
                "datetime",
                "task_history_id",
                "working_time",
            ]
        ],
        output=output,
        to_csv_kwargs=to_csv_kwargs,
    )


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    FindBreakError(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    parser.add_argument("--task_history_time_threshold", type=int, default=180, help="1履歴、何分以上を検知対象とするか。")
    parser.add_argument("--import_file_path", type=str, help="importするタスク履歴イベント全件ファイル,指定しない場合はタスク履歴イベント全件を新規取得する")

    argument_parser.add_output()
    parser.add_argument(
        "-p",
        "--project_id",
        type=str,
        required=True,
        nargs="+",
        help="対象のプロジェクトのproject_idを指定します。複数指定可、但しtask_idを指定した場合は1つしか指定できません。"
        "`file://`を先頭に付けると、project_idの一覧が記載されたファイルを指定できます。",
    )
    parser.add_argument(
        "-t",
        "--task_id",
        type=str,
        nargs="+",
        help="対象のプロジェクトのtask_idを指定します。複数指定可、但しtimeを指定した場合は1つしか指定できません。"
        "`file://`を先頭に付けると、task_idの一覧が記載されたファイルを指定できます。",
    )
    parser.add_argument(
        "--time",
        type=str,
        nargs="+",
        help="検索対象の時間を指定します。(%%Y/%%m/%%d %%H:%%i:%%s)",
    )
    parser.add_argument("--add", action="store_true", help="出力する際に追記で書き込む")

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "found_break_error"
    subcommand_help = "不当に長い作業時間の作業履歴を出力します"
    description = "自動休憩が作動せず不当に長い作業時間になっている履歴を出力します。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



import argparse

import annofabcli
import annofabcli.common.cli
import annofabcli.job.delete_job
import annofabcli.job.list_generated_task_history
import annofabcli.job.list_job
import annofabcli.job.list_last_job
import annofabcli.job.wait_job


def parse_args(parser: argparse.ArgumentParser):

    subparsers = parser.add_subparsers(dest="subcommand_name")

    # サブコマンドの定義

    annofabcli.job.delete_job.add_parser(subparsers)
    annofabcli.job.list_generated_task_history.add_parser(subparsers)
    annofabcli.job.list_job.add_parser(subparsers)
    annofabcli.job.list_last_job.add_parser(subparsers)
    annofabcli.job.wait_job.add_parser(subparsers)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "job"
    subcommand_help = "ジョブ関係のサブコマンド"
    description = "ジョブ関係のサブコマンド"

    parser = annofabcli.common.cli.add_parser(
        subparsers, subcommand_name, subcommand_help, description, is_subcommand=False
    )
    parse_args(parser)



import argparse
import json
import logging
from enum import Enum
from typing import List

import pandas
from annofabapi.models import ProjectMemberRole, Task, TaskPhase, TaskStatus

import annofabcli
import annofabcli.common.cli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import (
    AbstractCommandLineInterface,
    ArgumentParser,
    build_annofabapi_resource_and_login,
    get_json_from_args,
    get_wait_options_from_args,
)
from annofabcli.common.dataclasses import WaitOptions
from annofabcli.common.download import DownloadingFile
from annofabcli.common.enums import FormatArgument

logger = logging.getLogger(__name__)

DEFAULT_WAIT_OPTIONS = WaitOptions(interval=60, max_tries=360)
DEFAULT_TASK_ID_DELIMITER = "_"


class TaskStatusForSummary(Enum):
    """
    TaskStatusのサマリー用（知りたい情報をstatusにしている）
    """

    ANNOTATION_NOT_STARTED = "annotation_not_started"
    """教師付未着手"""
    INSPECTION_NOT_STARTED = "inspection_not_started"
    """検査未着手"""
    ACCEPTANCE_NOT_STARTED = "acceptance_not_started"
    """受入未着手"""

    WORKING = "working"
    BREAK = "break"
    ON_HOLD = "on_hold"
    COMPLETE = "complete"

    @staticmethod
    def from_task(task: Task) -> "TaskStatusForSummary":
        status = task["status"]
        if status == TaskStatus.NOT_STARTED.value:
            phase = task["phase"]
            if phase == TaskPhase.ANNOTATION.value:
                return TaskStatusForSummary.ANNOTATION_NOT_STARTED
            elif phase == TaskPhase.INSPECTION.value:
                return TaskStatusForSummary.INSPECTION_NOT_STARTED
            elif phase == TaskPhase.ACCEPTANCE.value:
                return TaskStatusForSummary.ACCEPTANCE_NOT_STARTED
            else:
                raise RuntimeError(f"phase={phase}が対象外です。")
        else:
            return TaskStatusForSummary(status)


def add_info_to_task(task: Task) -> Task:
    task["status_for_summary"] = TaskStatusForSummary.from_task(task).value
    return task


def create_task_count_summary_df(task_list: List[Task]) -> pandas.DataFrame:
    """
    タスク数の集計結果が格納されたDataFrameを取得する。

    Args:
        task_list:

    Returns:

    """

    def add_columns_if_not_exists(df: pandas.DataFrame, column: str):
        if column not in df.columns:
            df[column] = 0

    df_task = pandas.DataFrame([add_info_to_task(t) for t in task_list])
    df_summary = df_task.pivot_table(
        values="task_id", index=["account_id"], columns=["status_for_summary"], aggfunc="count", fill_value=0
    ).reset_index()
    for status in TaskStatusForSummary:
        add_columns_if_not_exists(df_summary, status.value)

    return df_summary


class SummarizeTaskCountByUser(AbstractCommandLineInterface):
    def create_user_df(self, project_id: str, account_id_list: List[str]) -> pandas.DataFrame:
        user_list = []
        for account_id in account_id_list:
            user = self.facade.get_project_member_from_account_id(project_id=project_id, account_id=account_id)
            if user is not None:
                user_list.append(user)
        return pandas.DataFrame(user_list, columns=["account_id", "user_id", "username", "biography"])

    def create_summary_df(self, project_id: str, task_list: List[Task]) -> pandas.DataFrame:
        df_task_count = create_task_count_summary_df(task_list)
        df_user = self.create_user_df(project_id, df_task_count["account_id"])
        if len(df_user) == 0:
            return pandas.DataFrame()

        df = pandas.merge(df_user, df_task_count, how="left", on=["account_id"])
        task_count_columns = [s.value for s in TaskStatusForSummary]
        df[task_count_columns] = df[task_count_columns].fillna(0)
        return df

    def print_summarize_df(self, df: pandas.DataFrame) -> None:
        columns = ["user_id", "username", "biography"] + [status.value for status in TaskStatusForSummary]
        target_df = df[columns].sort_values("user_id")
        annofabcli.utils.print_according_to_format(
            target_df,
            arg_format=FormatArgument(FormatArgument.CSV),
            output=self.output,
            csv_format=self.csv_format,
        )

    def main(self):
        args = self.args
        project_id = args.project_id
        super().validate_project(project_id, [ProjectMemberRole.OWNER, ProjectMemberRole.TRAINING_DATA_USER])

        if args.task_json is not None:
            task_json_path = args.task_json
        else:
            wait_options = get_wait_options_from_args(get_json_from_args(args.wait_options), DEFAULT_WAIT_OPTIONS)
            cache_dir = annofabcli.utils.get_cache_dir()
            task_json_path = cache_dir / f"{project_id}-task.json"

            downloading_obj = DownloadingFile(self.service)
            downloading_obj.download_task_json(
                project_id, dest_path=str(task_json_path), is_latest=args.latest, wait_options=wait_options
            )

        with open(task_json_path, encoding="utf-8") as f:
            task_list = json.load(f)

        df = self.create_summary_df(project_id, task_list)
        if len(df) > 0:
            self.print_summarize_df(df)
        else:
            logger.error(f"出力対象データが0件のため、出力しません。")


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()
    parser.add_argument(
        "--task_json",
        type=str,
        help="タスク情報が記載されたJSONファイルのパスを指定してます。JSONファイルは`$ annofabcli project download task`コマンドで取得できます。"
        "指定しない場合は、AnnoFabからタスク全件ファイルをダウンロードします。",
    )

    parser.add_argument(
        "--latest", action="store_true", help="最新のタスク一覧ファイルを参照します。このオプションを指定すると、タスク一覧ファイルを更新するのに数分待ちます。"
    )

    parser.add_argument(
        "--wait_options",
        type=str,
        help="タスク一覧ファイルの更新が完了するまで待つ際のオプションを、JSON形式で指定してください。"
        "`file://`を先頭に付けるとjsonファイルを指定できます。"
        'デフォルは`{"interval":60, "max_tries":360}` です。'
        "`interval`:完了したかを問い合わせる間隔[秒], "
        "`max_tires`:完了したかの問い合わせを最大何回行うか。",
    )

    argument_parser.add_csv_format()
    argument_parser.add_output()

    parser.set_defaults(subcommand_func=main)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    SummarizeTaskCountByUser(service, facade, args).main()


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "summarize_task_count_by_user"
    subcommand_help = "ユーザごとに、担当しているタスク数を出力します。"
    description = "ユーザごとに、担当しているタスク数をCSV形式で出力します。"
    epilog = "アノテーションユーザまたはオーナロールを持つユーザで実行してください。"
    parser = annofabcli.common.cli.add_parser(
        subparsers, subcommand_name, subcommand_help, description=description, epilog=epilog
    )
    parse_args(parser)



import argparse
import logging
import multiprocessing
import sys
from functools import partial
from typing import Any, Dict, List, Optional, Tuple

import annofabapi
from annofabapi.models import ProjectMemberRole

import annofabcli
import annofabcli.common.cli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import (
    AbstracCommandCinfirmInterface,
    AbstractCommandLineInterface,
    ArgumentParser,
    build_annofabapi_resource_and_login,
)

logger = logging.getLogger(__name__)


class UpdateMetadataMain(AbstracCommandCinfirmInterface):
    def __init__(self, service: annofabapi.Resource, all_yes: bool = False):
        self.service = service
        AbstracCommandCinfirmInterface.__init__(self, all_yes)

    def set_metadata_to_input_data(
        self,
        project_id: str,
        input_data_id: str,
        metadata: Dict[str, Any],
        overwrite_metadata: bool = False,
        input_data_index: Optional[int] = None,
    ) -> bool:
        logging_prefix = f"{input_data_index+1} 件目" if input_data_index is not None else ""

        input_data = self.service.wrapper.get_input_data_or_none(project_id, input_data_id)
        if input_data is None:
            logger.warning(f"{logging_prefix} 入力データは存在しないのでスキップします。input_data_id={input_data_id}")
            return False

        logger.debug(
            f"{logging_prefix} input_data_id={input_data['input_data_id']}, "
            f"input_data_name={input_data['input_data_name']}"
        )
        if not self.confirm_processing(f"入力データのメタデータを更新しますか？ input_data_id={input_data['input_data_id']}"):
            return False

        input_data["last_updated_datetime"] = input_data["updated_datetime"]
        if overwrite_metadata:
            input_data["metadata"] = metadata
        else:
            input_data["metadata"].update(metadata)

        self.service.api.put_input_data(project_id, input_data_id, request_body=input_data)
        logger.debug(f"{logging_prefix} 入力データを更新しました。input_data_id={input_data['input_data_id']}")
        return True

    def set_metadata_to_input_data_wrapper(
        self, tpl: Tuple[int, str], project_id: str, metadata: Dict[str, Any], overwrite_metadata: bool = False
    ):
        input_data_index, input_data_id = tpl
        return self.set_metadata_to_input_data(
            project_id,
            input_data_id,
            metadata=metadata,
            overwrite_metadata=overwrite_metadata,
            input_data_index=input_data_index,
        )

    def update_metadata_of_input_data(
        self,
        project_id: str,
        input_data_id_list: List[str],
        metadata: Dict[str, Any],
        overwrite_metadata: bool = False,
        parallelism: Optional[int] = None,
    ):
        if overwrite_metadata:
            logger.info(f"{len(input_data_id_list)} 件の入力データのmetadataを、{metadata} に変更します（上書き）。")
        else:
            logger.info(f"{len(input_data_id_list)} 件の入力データのmetadataに、{metadata} を追加します。")

        success_count = 0

        if parallelism is not None:
            partial_func = partial(
                self.set_metadata_to_input_data_wrapper,
                project_id=project_id,
                metadata=metadata,
                overwrite_metadata=overwrite_metadata,
            )
            with multiprocessing.Pool(parallelism) as pool:
                result_bool_list = pool.map(partial_func, enumerate(input_data_id_list))
                success_count = len([e for e in result_bool_list if e])

        else:
            # 逐次処理
            for input_data_index, input_data_id in enumerate(input_data_id_list):
                result = self.set_metadata_to_input_data(
                    project_id,
                    input_data_id,
                    metadata=metadata,
                    overwrite_metadata=overwrite_metadata,
                    input_data_index=input_data_index,
                )
                if result:
                    success_count += 1

        logger.info(f"{success_count} / {len(input_data_id_list)} 件の入力データのmetadataを変更しました。")


class UpdateMetadata(AbstractCommandLineInterface):
    @staticmethod
    def validate(args: argparse.Namespace) -> bool:
        COMMON_MESSAGE = "annofabcli input_data update_metadata: error:"

        if args.parallelism is not None and not args.yes:
            print(
                f"{COMMON_MESSAGE} argument --parallelism: '--parallelism'を指定するときは、必ず'--yes'を指定してください。",
                file=sys.stderr,
            )
            return False

        return True

    def main(self):
        args = self.args
        if not self.validate(args):
            return

        input_data_id_list = annofabcli.common.cli.get_list_from_args(args.input_data_id)
        metadata = annofabcli.common.cli.get_json_from_args(args.metadata)
        super().validate_project(args.project_id, [ProjectMemberRole.OWNER])
        main_obj = UpdateMetadataMain(self.service, all_yes=args.yes)
        main_obj.update_metadata_of_input_data(
            args.project_id,
            input_data_id_list,
            metadata,
            overwrite_metadata=args.overwrite,
            parallelism=args.parallelism,
        )


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    UpdateMetadata(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)
    argument_parser.add_project_id()
    argument_parser.add_input_data_id(required=True)

    parser.add_argument(
        "--metadata",
        required=True,
        type=str,
        help="入力データに設定する`metadata`をJSON形式で指定してください。メタデータの値は文字列です。" "`file://`を先頭に付けると、JSON形式のファイルを指定できます。",
    )

    parser.add_argument(
        "--overwrite",
        action="store_true",
        help="指定した場合、メタデータを上書きして更新します（すでに設定されているメタデータは削除されます）。指定しない場合、`--metadata`に指定されたキーのみ更新されます。",
    )

    parser.add_argument(
        "--parallelism", type=int, help="使用するプロセス数（並列度）を指定してください。指定する場合は必ず'--yes'を指定してください。指定しない場合は、逐次的に処理します。"
    )

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "update_metadata"
    subcommand_help = "入力データのメタデータを更新します。"
    description = "入力データのメタデータを更新します。"
    epilog = "オーナロールを持つユーザで実行してください。"
    parser = annofabcli.common.cli.add_parser(
        subparsers, subcommand_name, subcommand_help, description=description, epilog=epilog
    )
    parse_args(parser)



import argparse
import logging
import mimetypes
from typing import Optional

import requests
from annofabapi.models import ProjectMemberRole
from pyquery import PyQuery

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, build_annofabapi_resource_and_login

logger = logging.getLogger(__name__)


class CopyInstruction(AbstractCommandLineInterface):
    """
    作業ガイドをコピーする。
    """

    def validate_projects(self, src_project_id: str, dest_project_id: str):
        """
        適切なRoleが付与されているかを確認する。

        Raises:
             AuthorizationError: 自分自身のRoleがいずれかのRoleにも合致しなければ、AuthorizationErrorが発生する。
        """
        super().validate_project(src_project_id, project_member_roles=None)
        super().validate_project(
            dest_project_id, project_member_roles=[ProjectMemberRole.ACCEPTER, ProjectMemberRole.OWNER]
        )

    @staticmethod
    def get_instruction_image_id_from_url(url: str) -> str:
        # URL Queryを除いたURLを取得する
        url_without_query = url.split("?")[0]
        return url_without_query.split("/")[-1]

    @staticmethod
    def _get_mime_type_from_filename(filename: Optional[str]) -> str:
        """
        ファイル名からMIME TYPEを取得する。
        """
        DEFAULT_MIME_TYPE = "application/octet-stream"
        if filename is None:
            return DEFAULT_MIME_TYPE

        content_type = mimetypes.guess_type(filename)[0]
        return content_type if content_type is not None else DEFAULT_MIME_TYPE

    def upload_instruction_image(self, src_project_id: str, dest_project_id: str, pq_img: PyQuery) -> Optional[str]:
        """
        コピー元の作業ガイド画像を、コピー先にアップロードする。
        Args:
            src_project_id:
            dest_project_id:
            pq_img:

        Returns:
            コピー先の作業ガイドのURL。コピーできなかった場合はNoneを返す。

        """

        src_instruction_image_url: str = pq_img.attr["src"]
        if src_instruction_image_url is None:
            logger.warning(f"{pq_img} にsrc属性がないのでスキップします。")
            return None

        logger.debug(f"コピー元プロジェクトの {src_instruction_image_url} を、コピー先プロジェクトにアップロードします。")
        instruction_image_id = self.get_instruction_image_id_from_url(src_instruction_image_url)

        # alt属性値にファイル名が設定されている
        content_type = self._get_mime_type_from_filename(pq_img.attr["alt"])

        try:
            response_image = self.service.api._request_get_with_cookie(src_project_id, src_instruction_image_url)
        except requests.exceptions.RequestException as e:
            logger.warning(f"コピー元の作業ガイド画像の取得に失敗しました。: {e}")
            return None

        dest_instruction_image_url = self.service.wrapper.upload_data_as_instruction_image(
            dest_project_id, instruction_image_id, data=response_image.content, content_type=content_type
        )
        return dest_instruction_image_url

    def put_instruction(self, project_id: str, instruction_html: str) -> None:
        old_instruction = self.service.wrapper.get_latest_instruction(project_id)
        request_body = {
            "html": instruction_html,
            "last_updated_datetime": old_instruction["last_updated_datetime"] if old_instruction is not None else None,
        }
        self.service.api.put_instruction(project_id, request_body=request_body)

    def register_instruction(self, src_project_id: str, dest_project_id: str, instruction_html: str) -> None:
        """
        作業ガイド用HTMLを、作業ガイドとして登録する。
        作業ガイドHTMLに記載されている画像はダウンロードする。

        Args:
            project_id:
            instruction_html:

        Returns:

        """

        pq_html = PyQuery(instruction_html)
        pq_img_list = pq_html("img")

        for img_elm in pq_img_list:
            pq_img = PyQuery(img_elm)

            dest_instruction_image_url = self.upload_instruction_image(src_project_id, dest_project_id, pq_img)
            if dest_instruction_image_url is not None:
                pq_img.attr["src"] = dest_instruction_image_url  # pylint: disable=unsupported-assignment-operation

        self.put_instruction(dest_project_id, str(pq_html))

    def copy_instruction(self, src_project_id: str, dest_project_id: str) -> None:
        self.validate_projects(src_project_id, dest_project_id)
        src_project_title = self.facade.get_project_title(src_project_id)
        dest_project_title = self.facade.get_project_title(dest_project_id)

        src_instruction = self.service.wrapper.get_latest_instruction(src_project_id)
        if src_instruction is None:
            logger.warning(f"コピー元プロジェクト '{src_project_title}' に作業ガイドが設定されていません。終了します。")
            return

        if not self.confirm_processing(f"'{src_project_title}' の作業ガイドを、'{dest_project_title}' にコピーしますか？"):
            return

        self.register_instruction(src_project_id, dest_project_id, instruction_html=src_instruction["html"])

    def main(self) -> None:
        args = self.args

        self.copy_instruction(src_project_id=args.src_project_id, dest_project_id=args.dest_project_id)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    CopyInstruction(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    parser.add_argument("src_project_id", type=str, help="コピー元のプロジェクトのproject_id")
    parser.add_argument("dest_project_id", type=str, help="コピー先のプロジェクトのproject_id")
    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "copy"
    subcommand_help = "作業ガイドをコピーします。"
    description = "作業ガイドを別プロジェクトにコピーします。"
    epilog = "コピー先のプロジェクトに対して、チェッカーまたはオーナロールを持つユーザで実行してください。"
    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)



"""
検査コメント一覧を出力する。
"""

import argparse
import logging
from typing import Callable, List, Optional

import annofabapi
import requests
from annofabapi.models import Inspection

import annofabcli
import annofabcli.common.cli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, ArgumentParser, build_annofabapi_resource_and_login
from annofabcli.common.enums import FormatArgument
from annofabcli.common.visualize import AddProps

logger = logging.getLogger(__name__)

FilterInspectionFunc = Callable[[Inspection], bool]


def create_filter_func(
    only_reply: bool,
    exclude_reply: bool,
) -> Callable[[Inspection], bool]:
    def filter_inspection(arg_inspection: Inspection) -> bool:  # pylint: disable=too-many-return-statements
        # 返信コメントを除く
        if only_reply:
            if arg_inspection["parent_inspection_id"] is None:
                return False

        if exclude_reply:
            if arg_inspection["parent_inspection_id"] is not None:
                return False

        return True

    return filter_inspection


class PrintInspections(AbstractCommandLineInterface):
    """
    検査コメント一覧を出力する。
    """

    def __init__(self, service: annofabapi.Resource, facade: AnnofabApiFacade, args: argparse.Namespace):
        super().__init__(service, facade, args)
        self.visualize = AddProps(self.service, args.project_id)

    def filter_inspection_list(
        self,
        inspection_list: List[Inspection],
        task_id_list: Optional[List[str]] = None,
        arg_filter_inspection: Optional[FilterInspectionFunc] = None,
    ) -> List[Inspection]:
        """
        引数の検査コメント一覧に`commenter_username`など、ユーザが知りたい情報を追加する。

        Args:
            inspection_list: 検査コメント一覧
            filter_inspection: 検索コメントを絞り込むための関数

        Returns:
            情報が追加された検査コメント一覧
        """

        def filter_task_id(e):
            if task_id_list is None or len(task_id_list) == 0:
                return True
            return e["task_id"] in task_id_list

        def filter_inspection(e):
            if arg_filter_inspection is None:
                return True
            return arg_filter_inspection(e)

        inspection_list = [e for e in inspection_list if filter_inspection(e) and filter_task_id(e)]
        return [self.visualize.add_properties_to_inspection(e) for e in inspection_list]

    def print_inspections(
        self,
        project_id: str,
        task_id_list: List[str],
        filter_inspection: Optional[FilterInspectionFunc] = None,
    ):
        """
        検査コメントを出力する

        Args:
            project_id: 対象のproject_id
            task_id_list: 受け入れ完了にするタスクのtask_idのList
            filter_inspection: 検索コメントを絞り込むための関数

        Returns:

        """

        inspection_list = self.get_inspections(
            project_id, task_id_list=task_id_list, filter_inspection=filter_inspection
        )

        logger.info(f"検査コメントの件数: {len(inspection_list)}")

        self.print_according_to_format(inspection_list)

    def get_inspections_by_input_data(self, project_id: str, task_id: str, input_data_id: str, input_data_index: int):
        """入力データごとに検査コメント一覧を取得する。

        Args:
            project_id:
            task_id:
            input_data_id:
            input_data_index: タスク内のinput_dataの番号

        Returns:
            対象の検査コメント一覧
        """

        detail = {"input_data_index": input_data_index}
        inspectins, _ = self.service.api.get_inspections(project_id, task_id, input_data_id)
        return [self.visualize.add_properties_to_inspection(e, detail) for e in inspectins]

    def get_inspections(
        self, project_id: str, task_id_list: List[str], filter_inspection: Optional[FilterInspectionFunc] = None
    ) -> List[Inspection]:
        """検査コメント一覧を取得する。

        Args:
            project_id:
            task_id_list:

        Returns:
            対象の検査コメント一覧
        """

        all_inspections: List[Inspection] = []
        for task_id in task_id_list:
            try:
                task, _ = self.service.api.get_task(project_id, task_id)
                input_data_id_list = task["input_data_id_list"]
                logger.info(f"タスク '{task_id}' に紐づく検査コメントを取得します。input_dataの個数 = {len(input_data_id_list)}")
                for input_data_index, input_data_id in enumerate(input_data_id_list):

                    inspections = self.get_inspections_by_input_data(
                        project_id, task_id, input_data_id, input_data_index
                    )

                    if filter_inspection is not None:
                        inspections = [e for e in inspections if filter_inspection(e)]

                    all_inspections.extend(inspections)

            except requests.HTTPError as e:
                logger.warning(e)
                logger.warning(f"タスク task_id = {task_id} の検査コメントを取得できなかった。")

        return all_inspections

    def main(self):
        args = self.args
        task_id_list = annofabcli.common.cli.get_list_from_args(args.task_id)

        filter_inspection = create_filter_func(only_reply=args.only_reply, exclude_reply=args.exclude_reply)
        self.print_inspections(
            args.project_id,
            task_id_list,
            filter_inspection=filter_inspection,
        )


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()
    argument_parser.add_task_id(
        required=True,
        help_message="対象のタスクのtask_idを指定します。　" "`file://`を先頭に付けると、task_idの一覧が記載されたファイルを指定できます。",
    )

    reply_comment_group = parser.add_mutually_exclusive_group()
    reply_comment_group.add_argument("--only_reply", action="store_true", help="返信コメントのみを出力する。")
    reply_comment_group.add_argument("--exclude_reply", action="store_true", help="返信コメントを除外して出力する。")

    argument_parser.add_format(
        choices=[
            FormatArgument.CSV,
            FormatArgument.JSON,
            FormatArgument.PRETTY_JSON,
            FormatArgument.INSPECTION_ID_LIST,
        ],
        default=FormatArgument.CSV,
    )
    argument_parser.add_output()
    argument_parser.add_csv_format()
    argument_parser.add_query()

    parser.set_defaults(subcommand_func=main)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    PrintInspections(service, facade, args).main()


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "list"

    subcommand_help = "検査コメント一覧を出力します。"

    description = "検査コメント一覧を出力します。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



import argparse

import annofabcli
import annofabcli.common.cli
import annofabcli.project_member.change_project_members
import annofabcli.project_member.copy_project_members
import annofabcli.project_member.delete_users
import annofabcli.project_member.invite_users
import annofabcli.project_member.list_users
import annofabcli.project_member.put_project_members


def parse_args(parser: argparse.ArgumentParser):

    subparsers = parser.add_subparsers(dest="subcommand_name")

    # サブコマンドの定義
    annofabcli.project_member.change_project_members.add_parser(subparsers)
    annofabcli.project_member.copy_project_members.add_parser(subparsers)
    annofabcli.project_member.delete_users.add_parser(subparsers)
    annofabcli.project_member.invite_users.add_parser(subparsers)
    annofabcli.project_member.list_users.add_parser(subparsers)
    annofabcli.project_member.put_project_members.add_parser(subparsers)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "project_member"
    subcommand_help = "プロジェクトメンバ関係のサブコマンド"
    description = "プロジェクトメンバ関係のサブコマンド"

    parser = annofabcli.common.cli.add_parser(
        subparsers, subcommand_name, subcommand_help, description, is_subcommand=False
    )
    parse_args(parser)



import argparse
import logging
from enum import Enum
from typing import Any, Callable

from annofabapi.models import JobStatus, JobType

import annofabcli
import annofabcli.common.cli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import (
    AbstractCommandLineInterface,
    build_annofabapi_resource_and_login,
    get_json_from_args,
    get_wait_options_from_args,
)
from annofabcli.common.dataclasses import WaitOptions

logger = logging.getLogger(__name__)

DEFAULT_WAIT_OPTIONS = WaitOptions(interval=60, max_tries=360)


class DownloadTarget(Enum):
    TASK = "task"
    INPUT_DATA = "input_data"
    INSPECTION_COMMENT = "inspection_comment"
    TASK_HISTORY = "task_history"
    TASK_HISTORY_EVENT = "task_history_event"
    SIMPLE_ANNOTATION = "simple_annotation"
    FULL_ANNOTATION = "full_annotation"


class Download(AbstractCommandLineInterface):
    def is_job_progress(self, project_id: str, job_type: JobType):
        job_list = self.service.api.get_project_job(project_id, query_params={"type": job_type.value})[0]["list"]
        if len(job_list) > 0:
            if job_list[0]["job_status"] == JobStatus.PROGRESS.value:
                return True

        return False

    def update_file_and_wait(
        self, project_id: str, job_type: JobType, update_func: Callable[[str], Any], wait_options: WaitOptions
    ) -> None:
        """
        最新化処理が完了するまで待つ。
        """
        MAX_WAIT_MINUTUE = wait_options.max_tries * wait_options.interval / 60

        if self.is_job_progress(project_id, job_type=job_type):
            logger.info(f"ダウンロード対象の最新化処理が既に実行されています。")
        else:
            logger.info(f"ダウンロード対象の最新化処理を実行します。")
            update_func(project_id)

        logger.info(f"ダウンロード対象の最新化処理が完了するまで、最大{MAX_WAIT_MINUTUE}分間待ちます。")
        result = self.service.wrapper.wait_for_completion(
            project_id,
            job_type=job_type,
            job_access_interval=wait_options.interval,
            max_job_access=wait_options.max_tries,
        )
        if result:
            logger.info(f"ダウンロード対象の最新化処理が完了しました。")
        else:
            logger.info(f"ダウンロードの対象の最新化に失敗したか、または {MAX_WAIT_MINUTUE} 分待っても最新化処理が完了しませんでした。")

    def download(self, target: DownloadTarget, project_id: str, output: str, latest: bool, wait_options: WaitOptions):
        project_title = self.facade.get_project_title(project_id)
        logger.info(f"{project_title} の {target.value} をダウンロードします。")

        if target == DownloadTarget.TASK:
            if latest:
                self.update_file_and_wait(
                    project_id, JobType.GEN_TASKS_LIST, self.service.api.post_project_tasks_update, wait_options
                )

            self.service.wrapper.download_project_tasks_url(project_id, output)

        elif target == DownloadTarget.INPUT_DATA:
            if latest:
                self.update_file_and_wait(
                    project_id, JobType.GEN_INPUTS_LIST, self.service.api.post_project_inputs_update, wait_options
                )

            self.service.wrapper.download_project_inputs_url(project_id, output)

        elif target == DownloadTarget.INSPECTION_COMMENT:
            self.service.wrapper.download_project_inspections_url(project_id, output)

        elif target == DownloadTarget.TASK_HISTORY:
            self.service.wrapper.download_project_task_histories_url(project_id, output)

        elif target == DownloadTarget.TASK_HISTORY_EVENT:
            self.service.wrapper.download_project_task_history_events_url(project_id, output)

        elif target in [DownloadTarget.SIMPLE_ANNOTATION, DownloadTarget.FULL_ANNOTATION, DownloadTarget.TASK]:
            if latest:
                self.update_file_and_wait(
                    project_id, JobType.GEN_ANNOTATION, self.service.api.post_annotation_archive_update, wait_options
                )

            if target == DownloadTarget.SIMPLE_ANNOTATION:
                self.service.wrapper.download_annotation_archive(project_id, output, v2=True)

            elif target == DownloadTarget.FULL_ANNOTATION:
                self.service.wrapper.download_full_annotation_archive(project_id, output)

        logger.info(f"ダウンロードが完了しました。output={output}")

    @staticmethod
    def validate(args: argparse.Namespace):
        download_target = DownloadTarget(args.target)
        if args.latest:
            if download_target not in [
                DownloadTarget.TASK,
                DownloadTarget.INPUT_DATA,
                DownloadTarget.SIMPLE_ANNOTATION,
                DownloadTarget.FULL_ANNOTATION,
            ]:
                logger.warning(
                    f"ダウンロード対象が'task', 'input_data', 'simple_annotation', 'full_annotation'以外では`--latest`オプションは無視されます。"
                )

        if download_target == DownloadTarget.FULL_ANNOTATION:
            logger.warning(f"ダウンロード対象`full_annotation`は非推奨です。いずれ廃止されます。")

        return True

    def main(self):
        args = self.args
        if not self.validate(args):
            return

        wait_options = get_wait_options_from_args(get_json_from_args(args.wait_options), DEFAULT_WAIT_OPTIONS)
        self.download(
            DownloadTarget(args.target),
            args.project_id,
            output=args.output,
            latest=args.latest,
            wait_options=wait_options,
        )


def main(args: argparse.Namespace):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    Download(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    target_choices = [e.value for e in DownloadTarget]

    parser.add_argument(
        "target",
        type=str,
        choices=target_choices,
        help="ダウンロード対象の項目を指定します。"
        "simple_annotation: シンプルアノテーションzip, "
        "full_annotation: フルアノテーションzip(非推奨), "
        "task: タスクjson, "
        "input_data: 入力データjson, "
        "inspection_comment: 検査コメントjson, "
        "task_history: タスク履歴json, "
        "task_history_event: タスク履歴イベントjson",
    )

    parser.add_argument("-p", "--project_id", type=str, required=True, help="対象のプロジェクトのproject_idを指定します。")

    parser.add_argument("-o", "--output", type=str, required=True, help="ダウンロード先を指定します。")

    parser.add_argument(
        "--latest",
        action="store_true",
        help="ダウンロード対象を最新化してから、ダウンロードします。アノテーションの最新化は5分以上かかる場合があります。"
        "ダウンロード対象が'task', 'input_data', 'simple_annotation', 'full_annotation'のときのみ、このオプションは有効です。",
    )

    parser.add_argument(
        "--wait_options",
        type=str,
        help="ダウンロード対象の最新化を待つときのオプションをJSON形式で指定してください。"
        "`file://`を先頭に付けるとjsonファイルを指定できます。"
        'デフォルとは`{"interval":60, "max_tries":360}` です。'
        "`interval`:最新化が完了したかを問い合わせる間隔[秒], "
        "`max_tires`:最新化が完了したかの問い合わせを最大何回行うか。",
    )

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "download"
    subcommand_help = "タスクや検査コメント、アノテーションなどをダウンロードします。"
    description = (
        "タスクや検査コメント、アノテーションなどをダウンロードします。" + "タスク、検査コメント、タスク履歴イベントは毎日AM 02:00 JSTに更新されます。"
        "アノテーションは毎日AM 03:00 JSTに更新されます。"
    )
    epilog = "オーナロールを持つユーザで実行してください。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)



import argparse
import copy
import logging
import uuid
from typing import Any, Dict, Optional

from annofabapi.models import JobType, OrganizationMemberRole, ProjectMemberRole

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import (
    AbstractCommandLineInterface,
    ArgumentParser,
    build_annofabapi_resource_and_login,
    get_json_from_args,
    get_wait_options_from_args,
)
from annofabcli.common.dataclasses import WaitOptions

DEFAULT_WAIT_OPTIONS = WaitOptions(interval=60, max_tries=360)

logger = logging.getLogger(__name__)


class CopyProject(AbstractCommandLineInterface):
    """
    プロジェクトをコピーする
    """

    def copy_project(
        self,
        src_project_id: str,
        dest_project_id: str,
        dest_title: str,
        wait_options: WaitOptions,
        dest_overview: Optional[str] = None,
        copy_options: Optional[Dict[str, bool]] = None,
        wait_for_completion: bool = False,
    ):
        """
        プロジェクトメンバを、別のプロジェクトにコピーする。

        Args:
            src_project_id: コピー元のproject_id
            dest_project_id: 新しいプロジェクトのproject_id
            dest_title: 新しいプロジェクトのタイトル
            wait_options: 待つときのオプション
            dest_overview: 新しいプロジェクトの概要
            copy_options: 各項目についてコピーするかどうかのオプション
            wait_for_completion: プロジェクトのコピーが完了するまで待つかかどうか
        """

        self.validate_project(
            src_project_id,
            project_member_roles=[ProjectMemberRole.OWNER],
            organization_member_roles=[OrganizationMemberRole.ADMINISTRATOR, OrganizationMemberRole.OWNER],
        )

        src_project_title = self.facade.get_project_title(src_project_id)

        if copy_options is not None:
            copy_target = [key.replace("copy_", "") for key in copy_options.keys() if copy_options[key]]
            logger.info(f"コピー対象: {str(copy_target)}")

        confirm_message = f"{src_project_title} ({src_project_id} を、{dest_title} ({dest_project_id}) にコピーしますか？"
        if not self.confirm_processing(confirm_message):
            return

        request_body: Dict[str, Any] = {}
        if copy_options is not None:
            request_body = copy.deepcopy(copy_options)

        request_body.update(
            {"dest_project_id": dest_project_id, "dest_title": dest_title, "dest_overview": dest_overview}
        )

        self.service.api.initiate_project_copy(src_project_id, request_body=request_body)
        logger.info(f"プロジェクトのコピーを実施しています。")

        if wait_for_completion:
            MAX_WAIT_MINUTUE = wait_options.max_tries * wait_options.interval / 60
            logger.info(f"最大{MAX_WAIT_MINUTUE}分間、コピーが完了するまで待ちます。")

            result = self.service.wrapper.wait_for_completion(
                src_project_id,
                job_type=JobType.COPY_PROJECT,
                job_access_interval=wait_options.interval,
                max_job_access=wait_options.max_tries,
            )
            if result:
                logger.info(f"プロジェクトのコピーが完了しました。")
            else:
                logger.info(f"プロジェクトのコピーは実行中 または 失敗しました。")

    def main(self):
        args = self.args
        dest_project_id = args.dest_project_id if args.dest_project_id is not None else str(uuid.uuid4())

        copy_option_kyes = [
            "copy_inputs",
            "copy_tasks",
            "copy_annotations",
            "copy_webhooks",
            "copy_supplementaly_data",
            "copy_instructions",
        ]
        copy_options: Dict[str, bool] = {}
        for key in copy_option_kyes:
            copy_options[key] = getattr(args, key)

        wait_options = get_wait_options_from_args(get_json_from_args(args.wait_options), DEFAULT_WAIT_OPTIONS)

        self.copy_project(
            args.project_id,
            dest_project_id=dest_project_id,
            dest_title=args.dest_title,
            dest_overview=args.dest_overview,
            copy_options=copy_options,
            wait_for_completion=args.wait,
            wait_options=wait_options,
        )


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    CopyProject(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id(help_message="コピー元のプロジェクトのproject_idを指定してください。")

    parser.add_argument("--dest_project_id", type=str, help="新しいプロジェクトのproject_idを指定してください。省略した場合は UUIDv4 フォーマットになります。")
    parser.add_argument("--dest_title", type=str, required=True, help="新しいプロジェクトのタイトルを指定してください。")
    parser.add_argument("--dest_overview", type=str, help="新しいプロジェクトの概要を指定してください。")

    parser.add_argument("--copy_inputs", action="store_true", help="「入力データ」をコピーするかどうかを指定します。")
    parser.add_argument("--copy_tasks", action="store_true", help="「タスク」をコピーするかどうかを指定します。")
    parser.add_argument("--copy_annotations", action="store_true", help="「アノテーション」をコピーするかどうかを指定します。")
    parser.add_argument("--copy_webhooks", action="store_true", help="「Webhook」をコピーするかどうかを指定します。")
    parser.add_argument("--copy_supplementaly_data", action="store_true", help="「補助情報」をコピーするかどうかを指定します。")
    parser.add_argument("--copy_instructions", action="store_true", help="「作業ガイド」をコピーするかどうかを指定します。")

    parser.add_argument("--wait", action="store_true", help="プロジェクトのコピーが完了するまで待ちます。")

    parser.add_argument(
        "--wait_options",
        type=str,
        help="プロジェクトのコピーが完了するまで待つ際のオプションをJSON形式で指定してください。"
        "`file://`を先頭に付けるとjsonファイルを指定できます。"
        'デフォルとは`{"interval":60, "max_tries":360}` です。'
        "`interval`:完了したかを問い合わせる間隔[秒], "
        "`max_tires`:完了したかの問い合わせを最大何回行うか。",
    )

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "copy"
    subcommand_help = "プロジェクトをコピーします。"
    description = "プロジェクトをコピーして（アノテーション仕様やメンバーを引き継いで）、新しいプロジェクトを作成します。"
    epilog = "コピー元のプロジェクトに対してオーナロール、組織に対して組織管理者、組織オーナを持つユーザで実行してください。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)



"""
プロジェクトのユーザを表示する。
"""
import argparse
import logging
from typing import Any, Dict, List, Optional

import annofabapi
from annofabapi.models import Task

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, ArgumentParser, build_annofabapi_resource_and_login
from annofabcli.common.enums import FormatArgument
from annofabcli.common.visualize import AddProps

logger = logging.getLogger(__name__)


class ListTasksMain:
    def __init__(self, service: annofabapi.Resource, project_id: str):
        self.service = service
        self.facade = AnnofabApiFacade(service)
        self.project_id = project_id
        self.visualize = AddProps(self.service, project_id)

    @annofabcli.utils.allow_404_error
    def get_task(self, project_id: str, task_id: str) -> Task:
        task, _ = self.service.api.get_task(project_id, task_id)
        return task

    def get_task_list_from_task_id(self, project_id: str, task_id_list: List[str]) -> List[Task]:
        task_list = []
        logger.debug(f"{len(task_id_list)}件のタスクを取得します。")
        for index, task_id in enumerate(task_id_list):
            if (index + 1) % 100 == 0:
                logger.debug(f"{index+1} 件のタスクを取得します。")

            task = self.get_task(project_id, task_id)
            if task is not None:
                task_list.append(self.visualize.add_properties_to_task(task))
            else:
                logger.warning(f"タスク '{task_id}' は見つかりませんでした。")

        return task_list

    def _modify_task_query(self, project_id: str, task_query: Dict[str, Any]) -> Dict[str, Any]:
        """
        タスク検索クエリを修正する。
        ``user_id`` から ``account_id`` に変換する。
        ``previous_user_id`` から ``previcous_account_id`` に変換する。
        ``page`` , ``limit``を削除」する

        Args:
            task_query: タスク検索クエリ（変更される）

        Returns:
            修正したタスク検索クエリ

        """

        def remove_key(arg_key: str):
            if arg_key in task_query:
                logger.info(f"タスク検索クエリから、`{arg_key}`　キーを削除しました。")
                task_query.pop(arg_key)

        remove_key("page")
        remove_key("limit")

        if "user_id" in task_query:
            user_id = task_query["user_id"]
            account_id = self.facade.get_account_id_from_user_id(project_id, user_id)
            if account_id is not None:
                task_query["account_id"] = account_id
            else:
                logger.warning(f"タスク検索クエリに含まれている user_id: {user_id} のユーザが見つかりませんでした。")

        if "previous_user_id" in task_query:
            previous_user_id = task_query["previous_user_id"]
            previous_account_id = self.facade.get_account_id_from_user_id(project_id, previous_user_id)
            if previous_account_id is not None:
                task_query["previous_account_id"] = previous_account_id
            else:
                logger.warning(f"タスク検索クエリに含まれている previous_user_id: {previous_user_id} のユーザが見つかりませんでした。")

        return task_query

    def get_task_list_with_api(
        self, project_id: str, task_query: Optional[Dict[str, Any]] = None, user_id_list: Optional[List[str]] = None
    ) -> List[Task]:
        """
        タスク一覧を取得する。

        Args:
            project_id:
            task_id_list:

        Returns:
            対象の検査コメント一覧
        """
        if task_query is not None:
            task_query = self._modify_task_query(project_id, task_query)
        else:
            task_query = {}

        if user_id_list is None:
            logger.debug(f"task_query: {task_query}")
            tasks = self.service.wrapper.get_all_tasks(project_id, query_params=task_query)
            if len(tasks) == 10000:
                logger.warning("タスク一覧は10,000件で打ち切られている可能性があります。")

        else:
            tasks = []
            for user_id in user_id_list:
                task_query["user_id"] = user_id
                task_query = self._modify_task_query(project_id, task_query)
                logger.debug(f"task_query: {task_query}")
                sub_tasks = self.service.wrapper.get_all_tasks(project_id, query_params=task_query)
                if len(sub_tasks) == 10000:
                    logger.warning(f"user_id={user_id}で絞り込んだタスク一覧は10,000件で打ち切られている可能性があります。")
                tasks.extend(sub_tasks)

        return [self.visualize.add_properties_to_task(e) for e in tasks]

    def get_task_list(
        self,
        project_id: str,
        task_id_list: Optional[List[str]] = None,
        task_query: Optional[Dict[str, Any]] = None,
        user_id_list: Optional[List[str]] = None,
    ) -> List[Task]:
        """


        Args:
            project_id: 対象のproject_id
            task_id_list: 対象のタスクのtask_id
            task_query: タスク検索クエリ
            user_id_list:

        Returns:

        """
        if task_id_list is not None:
            task_list = self.get_task_list_from_task_id(project_id, task_id_list=task_id_list)
        else:
            task_list = self.get_task_list_with_api(project_id, task_query=task_query, user_id_list=user_id_list)

        return task_list


class ListTasks(AbstractCommandLineInterface):
    """
    タスクの一覧を表示する
    """

    def __init__(self, service: annofabapi.Resource, facade: AnnofabApiFacade, args: argparse.Namespace):
        super().__init__(service, facade, args)
        self.visualize = AddProps(self.service, args.project_id)

    def main(self):
        args = self.args

        task_id_list = annofabcli.common.cli.get_list_from_args(args.task_id)
        if len(task_id_list) == 0:
            task_id_list = None

        user_id_list = annofabcli.common.cli.get_list_from_args(args.user_id)
        if len(user_id_list) == 0:
            user_id_list = None

        task_query = annofabcli.common.cli.get_json_from_args(args.task_query)

        project_id = args.project_id
        super().validate_project(project_id, project_member_roles=None)

        main_obj = ListTasksMain(self.service, project_id=project_id)
        task_list = main_obj.get_task_list(
            project_id=project_id,
            task_id_list=task_id_list,
            task_query=task_query,
            user_id_list=user_id_list,
        )
        logger.debug(f"タスク一覧の件数: {len(task_list)}")

        if len(task_list) > 0:
            self.print_according_to_format(task_list)
        else:
            logger.info(f"タスク一覧の件数が0件のため、出力しません。")


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ListTasks(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()

    query_group = parser.add_mutually_exclusive_group()

    # タスク検索クエリ
    query_group.add_argument(
        "-tq",
        "--task_query",
        type=str,
        help="タスクの検索クエリをJSON形式で指定します。指定しない場合は、すべてのタスクを取得します。"
        "`file://`を先頭に付けると、JSON形式のファイルを指定できます。"
        "クエリのフォーマットは、[getTasks API](https://annofab.com/docs/api/#operation/getTasks)のクエリパラメータと同じです。"
        "さらに追加で、`user_id`, `previous_user_id` キーも指定できます。"
        "ただし `page`, `limit`キーは指定できません。",
    )

    query_group.add_argument(
        "-t",
        "--task_id",
        type=str,
        nargs="+",
        help="対象のタスクのtask_idを指定します。`--task_query`引数とは同時に指定できません。" "`file://`を先頭に付けると、task_idの一覧が記載されたファイルを指定できます。",
    )

    parser.add_argument(
        "-u",
        "--user_id",
        type=str,
        nargs="+",
        help="絞り込み対象である担当者のuser_idを指定します。" "`file://`を先頭に付けると、task_idの一覧が記載されたファイルを指定できます。",
    )

    argument_parser.add_format(
        choices=[FormatArgument.CSV, FormatArgument.JSON, FormatArgument.PRETTY_JSON, FormatArgument.TASK_ID_LIST],
        default=FormatArgument.CSV,
    )
    argument_parser.add_output()
    argument_parser.add_csv_format()

    argument_parser.add_query()
    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "list"
    subcommand_help = "タスク一覧を出力します。"
    description = "タスク一覧を出力します。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



import argparse
import logging
import multiprocessing
from functools import partial
from typing import List, Optional

import annofabapi
import dateutil
from annofabapi.models import JobStatus, JobType, ProjectMemberRole

import annofabcli
import annofabcli.common.cli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import (
    AbstractCommandLineInterface,
    build_annofabapi_resource_and_login,
    get_json_from_args,
    get_wait_options_from_args,
)
from annofabcli.common.dataclasses import WaitOptions

logger = logging.getLogger(__name__)

DEFAULT_WAIT_OPTIONS = WaitOptions(interval=300, max_tries=72)


class SubUpdateAnnotationZip:
    """
    `AbstractCommandLineInterface`を継承したクラスだと、`multiprocessing.Pool`を実行したときに
    `AttributeError: Can't pickle local object 'ArgumentParser.__init__.<locals>.identity'`というエラーが発生したので、
    `ArgumentParser`を除いたクラスを作成した。
    """

    def __init__(self, service: annofabapi.Resource, facade: AnnofabApiFacade):
        self.service = service
        self.facade = facade

    def _should_update_annotation_zip(self, project_id: str) -> bool:
        """
        アノテーションzipを更新する必要があるかどうか

        Args:
            project_id:

        Returns:
            True:アノテーションzipを更新する必要がある。

        """
        project, _ = self.service.api.get_project(project_id)
        last_tasks_updated_datetime = project["summary"]["last_tasks_updated_datetime"]
        logger.debug(f"project_id={project_id}: タスクの最終更新日時={last_tasks_updated_datetime}")
        if last_tasks_updated_datetime is None:
            logger.debug(f"project_id={project_id}: タスクがまだ作成されていないので、アノテーションzipを更新する必要はありません。")
            return False

        annotation_specs_history = self.service.api.get_annotation_specs_histories(project_id)[0]
        annotation_specs_updated_datetime = annotation_specs_history[-1]["updated_datetime"]
        logger.debug(f"project_id={project_id}: アノテーション仕様の最終更新日時={annotation_specs_updated_datetime}")

        job_list = self.service.api.get_project_job(
            project_id, query_params={"type": JobType.GEN_ANNOTATION.value, "limit": 1}
        )[0]["list"]

        if len(job_list) == 0:
            return True

        job = job_list[0]
        logger.debug(
            f"project_id={project_id}: 最後にアノテーションzipを更新しときの情報: "
            f"最終更新日時={job['updated_datetime']}, job_status={job['job_status']}"
        )
        job_status = JobStatus(job["job_status"])
        if job_status == JobStatus.SUCCEEDED:
            if dateutil.parser.parse(job["updated_datetime"]) < dateutil.parser.parse(last_tasks_updated_datetime):
                return True

            elif dateutil.parser.parse(job["updated_datetime"]) < dateutil.parser.parse(
                annotation_specs_updated_datetime
            ):
                return True
            else:
                logger.debug(
                    f"project_id={project_id}: タスクの最終更新日時 or アノテーション仕様の最終更新日時が、"
                    f"アノテーションzipの最終講師日時より新しいため、アノテーションzipを更新する必要はありません。"
                )
                return False
        else:
            return True

    def _wait_for_completion_updated_annotation(self, project_id: str, wait_options: Optional[WaitOptions] = None):
        """
        アノテーションzipの更新が完了するまで待ちます。
        """
        if wait_options is None:
            wait_options = DEFAULT_WAIT_OPTIONS

        MAX_WAIT_MINUTU = wait_options.max_tries * wait_options.interval / 60
        result = self.service.wrapper.wait_for_completion(
            project_id,
            job_type=JobType.GEN_ANNOTATION,
            job_access_interval=wait_options.interval,
            max_job_access=wait_options.max_tries,
        )
        if result:
            logger.info(f"project_id={project_id}: アノテーションzipの更新が完了しました。")
        else:
            logger.info(f"project_id={project_id}: アノテーションzipの更新に失敗、または{MAX_WAIT_MINUTU}分待ってもアノテーションzipの更新が終了しませんでした。")
        return result

    def _update_annotation_zip_for_project(self, project_id: str) -> None:
        job_list = self.service.api.get_project_job(
            project_id, query_params={"type": JobType.GEN_ANNOTATION.value, "limit": 1}
        )[0]["list"]
        if len(job_list) > 0:
            job = job_list[0]
            if job["job_status"] == JobStatus.PROGRESS.value:
                logger.info(f"project_id={project_id}: アノテーションzipの更新処理が既に実行されています。")
                return

        self.service.api.post_annotation_archive_update(project_id)
        logger.info(f"project_id={project_id}: アノテーションzipの更新処理が開始されました。")

    def execute_for_project(
        self, project_id: str, force: bool = False, wait: bool = False, wait_options: Optional[WaitOptions] = None
    ) -> None:
        """
        1つのプロジェクトに対して、アノテーションzipを更新して、必要なら更新が完了するまで待ちます。
        """
        project = self.service.wrapper.get_project_or_none(project_id)
        if project is None:
            logger.warning(f"project_id={project_id}: プロジェクトにアクセスできません。")
            return

        logger.debug(f"project_id={project_id}: project_title='{project['title']}'")
        if not self.facade.contains_any_project_member_role(
            project_id, [ProjectMemberRole.OWNER, ProjectMemberRole.TRAINING_DATA_USER]
        ):
            logger.warning(f"project_id={project_id}: オーナロールまたはアノテーションユーザロールでないため、アノテーションzipを更新できません。")
            return

        if not force:
            should_update = self._should_update_annotation_zip(project_id)
        else:
            should_update = True

        if should_update:
            self._update_annotation_zip_for_project(project_id)
            if wait:
                self._wait_for_completion_updated_annotation(project_id, wait_options)
        else:
            logger.info(f"project_id={project_id}: アノテーションzipを更新する必要がないので、何も処理しません。")


class UpdateAnnotationZip(AbstractCommandLineInterface):
    def update_annotation_zip(
        self,
        project_id_list: List[str],
        force: bool = False,
        wait: bool = False,
        wait_options: Optional[WaitOptions] = None,
        parallelism: Optional[int] = None,
    ) -> None:
        """
        複数プロジェクトに対して、アノテーションzipを更新する。

        Args:
            project_id:
            force: アノテーションzipを更新する必要がなくても、常に更新する。
            wait: アノテーションzipの更新が完了するまで待つかどうか
            wait_options: アノテーションzipの更新が完了するまで待つときのオプション
            parallelism: 並列度
        """
        obj = SubUpdateAnnotationZip(service=self.service, facade=self.facade)
        processes = parallelism if parallelism is not None else len(project_id_list)
        # project_idごとに並列で処理します
        partial_func = partial(obj.execute_for_project, force=force, wait=wait, wait_options=wait_options)
        with multiprocessing.Pool(processes) as pool:
            pool.map(partial_func, project_id_list)

    def main(self):
        args = self.args

        project_id_list = annofabcli.common.cli.get_list_from_args(args.project_id)

        wait_options = get_wait_options_from_args(get_json_from_args(args.wait_options), DEFAULT_WAIT_OPTIONS)
        self.update_annotation_zip(
            project_id_list=project_id_list,
            force=args.force,
            wait=args.wait,
            wait_options=wait_options,
        )


def main(args: argparse.Namespace):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    UpdateAnnotationZip(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    parser.add_argument(
        "-p",
        "--project_id",
        type=str,
        nargs="+",
        required=True,
        help="対象のプロジェクトのproject_idを指定します。`file://`を先頭に付けると、project_idの一覧が記載されたファイルを指定できます。",
    )

    parser.add_argument(
        "--force", action="store_true", help="アノテーションzipを常に更新します。指定しない場合は、アノテーションzipを更新する必要がなければ更新しません。"
    )

    parser.add_argument("--wait", action="store_true", help="アノテーションzipの更新が完了するまで待ちます。")

    parser.add_argument(
        "--wait_options",
        type=str,
        help="アノテーションzipの最新化が完了するまで待つ際のオプションを、JSON形式で指定してください。"
        "`file://`を先頭に付けるとjsonファイルを指定できます。"
        'デフォルとは`{"interval":300, "max_tries":72}` です。'
        "`interval`:完了したかを問い合わせる間隔[秒], "
        "`max_tires`:完了したかの問い合わせを最大何回行うか。",
    )

    parser.add_argument("--parallelism", type=int, help="並列度。指定しない場合は、project_idの個数が並列度になります。")
    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "update_annotation_zip"
    subcommand_help = "アノテーションzipを更新します。"
    description = "アノテーションzipを更新します。"
    epilog = "対象プロジェクトに対して、オーナまたはアノテーションユーザロールを持つユーザで実行してください。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)



import argparse
import logging
from pathlib import Path

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, build_annofabapi_resource_and_login
from annofabcli.common.utils import read_multiheader_csv
from annofabcli.statistics.scatter import Scatter

logger = logging.getLogger(__name__)


def write_performance_scatter_per_user(csv: Path, output_dir: Path) -> None:
    df = read_multiheader_csv(str(csv))
    scatter_obj = Scatter(str(output_dir))
    scatter_obj.write_scatter_for_productivity_by_monitored_worktime(df)
    scatter_obj.write_scatter_for_productivity_by_actual_worktime(df)
    scatter_obj.write_scatter_for_quality(df)
    scatter_obj.write_scatter_for_productivity_by_actual_worktime_and_quality(df)


class WriteScatterPerUser(AbstractCommandLineInterface):
    def main(self):
        args = self.args
        write_performance_scatter_per_user(csv=args.csv, output_dir=args.output_dir)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    WriteScatterPerUser(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--csv",
        type=Path,
        required=True,
        help=("CSVファイルのパスを指定してください。" "CSVは、'statistics visualize'コマンドの出力結果である'メンバごとの生産性と品質.csv'と同じフォーマットです。"),
    )
    parser.add_argument("-o", "--output_dir", type=Path, required=True, help="出力ディレクトリのパス")

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "write_performance_scatter_per_user"
    subcommand_help = "CSVからユーザごとにプロットした散布図を出力します。"
    description = "CSVからユーザごとにプロットした散布図を出力します。"
    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



from dataclasses import dataclass
from typing import Any, Dict

from dataclasses_json import DataClassJsonMixin


@dataclass(frozen=True)
class WaitOptions(DataClassJsonMixin):
    """
    最新化ジョブが完了するまで待つときのオプション
    """

    interval: int
    """ジョブにアクセスする間隔[秒]"""

    max_tries: int
    """最大ジョブに何回アクセスするか"""


@dataclass
class SimpleAnnotationDetail4Import(DataClassJsonMixin):
    """
    アノテーションインポート用の　``SimpleAnnotationDetail`` クラス。
    """

    label: str
    """アノテーション仕様のラベル名(英語)です。 """

    data: Dict[str, Any]
    """"""

    attributes: Dict[str, Any]
    """キーに属性の名前、値に各属性の値が入った辞書構造です。 """



import asyncio
import logging.config
import warnings
from functools import partial
from typing import Optional

import annofabapi
import requests
from annofabapi.models import JobType

from annofabcli.common.dataclasses import WaitOptions
from annofabcli.common.exceptions import DownloadingFileNotFoundError, UpdatedFileForDownloadingError

logger = logging.getLogger(__name__)

DOWNLOADING_FILETYPE_DICT = {
    JobType.GEN_TASKS_LIST: "タスク全件ファイル",
    JobType.GEN_INPUTS_LIST: "入力データ全件ファイル",
    JobType.GEN_ANNOTATION: "アノテーションzip",
}

DEFAULT_WAIT_OPTIONS = WaitOptions(interval=60, max_tries=360)


def _get_annofab_error_message(http_error: requests.HTTPError) -> Optional[str]:
    obj = http_error.response.json()
    errors = obj.get("errors")
    if errors is None:
        return None
    return errors[0].get("message")


class DownloadingFile:
    """"""

    def __init__(self, service: annofabapi.Resource):
        self.service = service

    @staticmethod
    def get_max_wait_minutes(wait_options: WaitOptions):
        return wait_options.max_tries * wait_options.interval / 60

    def _wait_for_completion(
        self,
        project_id: str,
        job_type: JobType,
        wait_options: Optional[WaitOptions] = None,
        job_id: Optional[str] = None,
    ):
        if wait_options is None:
            wait_options = DEFAULT_WAIT_OPTIONS

        max_wait_minutues = self.get_max_wait_minutes(wait_options)
        filetype = DOWNLOADING_FILETYPE_DICT[job_type]
        logger.info(f"{filetype}の更新処理が完了するまで、最大{max_wait_minutues}分間待ちます。job_id={job_id}")
        result = self.service.wrapper.wait_for_completion(
            project_id,
            job_type=job_type,
            job_access_interval=wait_options.interval,
            max_job_access=wait_options.max_tries,
        )
        if not result:
            raise UpdatedFileForDownloadingError(f"{filetype}の更新処理が{max_wait_minutues}分以内に完了しない、または更新処理に失敗しました。")

    async def download_annotation_zip_with_async(
        self, project_id: str, dest_path: str, is_latest: bool = False, wait_options: Optional[WaitOptions] = None
    ):
        loop = asyncio.get_event_loop()
        partial_func = partial(self.download_annotation_zip, project_id, dest_path, is_latest, wait_options)
        result = await loop.run_in_executor(None, partial_func)
        return result

    def download_annotation_zip(
        self, project_id: str, dest_path: str, is_latest: bool = False, wait_options: Optional[WaitOptions] = None
    ):
        logger.debug(f"アノテーションzipをダウンロードします。path={dest_path}")
        if is_latest:
            self.wait_until_updated_annotation_zip(project_id, wait_options)
            self.service.wrapper.download_annotation_archive(project_id, dest_path, v2=True)

        else:
            try:
                self.service.wrapper.download_annotation_archive(project_id, dest_path, v2=True)
            except requests.HTTPError as e:
                if e.response.status_code == requests.codes.not_found:
                    logger.info(f"アノテーションzipが存在しなかったので、アノテーションzipファイルの更新処理を実行します。")
                    self.wait_until_updated_annotation_zip(project_id, wait_options)
                    self.service.wrapper.download_annotation_archive(project_id, dest_path, v2=True)
                else:
                    raise e

    def wait_until_updated_annotation_zip(self, project_id: str, wait_options: Optional[WaitOptions] = None):
        job_id = None
        try:
            job = self.service.api.post_annotation_archive_update(project_id, query_params={"v": "2"})[0]["job"]
            job_id = job["job_id"]
        except requests.HTTPError as e:
            # すでにジョブが進行中の場合は、無視する
            if e.response.status_code == requests.codes.conflict:
                logger.warning(f"別のバックグラウンドジョブが既に実行されているので、更新処理を無視します。")
                logger.warning(f"{_get_annofab_error_message(e)}")
            else:
                raise e

        self._wait_for_completion(project_id, job_type=JobType.GEN_ANNOTATION, wait_options=wait_options, job_id=job_id)

    async def download_input_data_json_with_async(
        self, project_id: str, dest_path: str, is_latest: bool = False, wait_options: Optional[WaitOptions] = None
    ):
        loop = asyncio.get_event_loop()
        partial_func = partial(self.download_input_data_json, project_id, dest_path, is_latest, wait_options)
        result = await loop.run_in_executor(None, partial_func)
        return result

    def download_input_data_json(
        self, project_id: str, dest_path: str, is_latest: bool = False, wait_options: Optional[WaitOptions] = None
    ):
        logger.debug(f"入力データ全件ファイルをダウンロードします。path={dest_path}")
        if is_latest:
            self.wait_until_updated_input_data_json(project_id, wait_options)
            self.service.wrapper.download_project_inputs_url(project_id, dest_path)

        else:
            try:
                self.service.wrapper.download_project_inputs_url(project_id, dest_path)
            except requests.HTTPError as e:
                if e.response.status_code == requests.codes.not_found:
                    logger.info(f"入力データ全件ファイルが存在しなかったので、入力データ全件ファイルの更新処理を実行します。")
                    self.wait_until_updated_input_data_json(project_id, wait_options)
                    self.service.wrapper.download_project_inputs_url(project_id, dest_path)
                else:
                    raise e

    def wait_until_updated_input_data_json(self, project_id: str, wait_options: Optional[WaitOptions] = None):
        job_id = None
        try:
            job = self.service.api.post_project_inputs_update(project_id)[0]["job"]
            job_id = job["job_id"]
        except requests.HTTPError as e:
            # すでにジョブが進行中の場合は、無視する
            if e.response.status_code == requests.codes.conflict:
                logger.warning(f"別のバックグラウンドジョブが既に実行されているので、更新処理を無視します。")
                logger.warning(f"{_get_annofab_error_message(e)}")
            else:
                raise e

        self._wait_for_completion(
            project_id, job_type=JobType.GEN_INPUTS_LIST, wait_options=wait_options, job_id=job_id
        )

    async def download_task_json_with_async(
        self, project_id: str, dest_path: str, is_latest: bool = False, wait_options: Optional[WaitOptions] = None
    ):
        loop = asyncio.get_event_loop()
        partial_func = partial(self.download_task_json, project_id, dest_path, is_latest, wait_options)
        result = await loop.run_in_executor(None, partial_func)
        return result

    def download_task_json(
        self, project_id: str, dest_path: str, is_latest: bool = False, wait_options: Optional[WaitOptions] = None
    ):
        logger.debug(f"タスク全件ファイルをダウンロードします。path={dest_path}")
        if is_latest:
            self.wait_until_updated_task_json(project_id, wait_options)
            self.service.wrapper.download_project_tasks_url(project_id, dest_path)

        else:
            try:
                self.service.wrapper.download_project_tasks_url(project_id, dest_path)
            except requests.HTTPError as e:
                if e.response.status_code == requests.codes.not_found:
                    logger.info(f"タスク全件ファイルが存在しなかったので、タスク全件ファイルの更新処理を実行します。")
                    self.wait_until_updated_task_json(project_id, wait_options)
                    self.service.wrapper.download_project_tasks_url(project_id, dest_path)
                else:
                    raise e

    def wait_until_updated_task_json(self, project_id: str, wait_options: Optional[WaitOptions] = None):
        job_id = None
        try:
            job = self.service.api.post_project_tasks_update(project_id, query_params={"v": "2"})[0]["job"]
            job_id = job["job_id"]
        except requests.HTTPError as e:
            # すでにジョブが進行中の場合は、無視する
            if e.response.status_code == requests.codes.conflict:
                logger.warning(f"別のバックグラウンドジョブが既に実行されているので、更新処理を無視します。")
                logger.warning(f"{_get_annofab_error_message(e)}")
            else:
                raise e

        self._wait_for_completion(project_id, job_type=JobType.GEN_TASKS_LIST, wait_options=wait_options, job_id=job_id)

    async def download_task_history_json_with_async(self, project_id: str, dest_path: str):
        """
        非同期でタスク履歴全件ファイルをダウンロードする。

        Raises:
            DownloadingFileNotFoundError:
        """
        return self.download_task_history_json(project_id, dest_path=dest_path)

    def download_task_history_json(self, project_id: str, dest_path: str):
        """
        タスク履歴全件ファイルをダウンロードする。

        Args:
            project_id:
            dest_path:

        Raises:
            DownloadingFileNotFoundError:
        """
        try:
            logger.debug(f"タスク履歴全件ファイルをダウンロードします。path={dest_path}")
            self.service.wrapper.download_project_task_histories_url(project_id, dest_path)
        except requests.HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                logger.info(f"タスク履歴全件ファイルが存在しません。")
                raise DownloadingFileNotFoundError("タスク履歴全件ファイルが存在しません。") from e
            raise e

    def download_task_history_event_json(self, project_id: str, dest_path: str):
        """
        タスク履歴イベント全件ファイルをダウンロードする。

        .. deprecated:: 0.21.1

        Args:
            project_id:
            dest_path:

        Raises:
            DownloadingFileNotFoundError:
        """
        try:
            logger.debug(f"タスク履歴イベント全件ファイルをダウンロードします。path={dest_path}")
            self.service.wrapper.download_project_task_history_events_url(project_id, dest_path)
        except requests.HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                logger.info(f"タスク履歴イベント全件ファイルが存在しません。")
                raise DownloadingFileNotFoundError("タスク履歴イベント全件ファイルが存在しません。") from e
            raise e

    async def download_task_history_event_json_with_async(self, project_id: str, dest_path: str):
        """
        非同期で検査コメント全件ファイルをダウンロードする。

        .. deprecated:: 0.21.1

        Raises:
            DownloadingFileNotFoundError:
        """
        warnings.warn("deprecated", DeprecationWarning)
        return self.download_task_history_event_json(project_id, dest_path=dest_path)

    async def download_inspection_json_with_async(self, project_id: str, dest_path: str):
        """
        非同期で検査コメント全件ファイルをダウンロードする。

        Raises:
            DownloadingFileNotFoundError:
        """

        return self.download_inspection_json(project_id, dest_path=dest_path)

    def download_inspection_json(self, project_id: str, dest_path: str):
        """
        検査コメント全件ファイルをダウンロードする。


        Raises:
            DownloadingFileNotFoundError:
        """
        try:
            logger.debug(f"検査コメント全件ファイルをダウンロードします。path={dest_path}")
            self.service.wrapper.download_project_inspections_url(project_id, dest_path)
        except requests.HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                logger.info(f"検査コメント全件ファイルが存在しません。")
                raise DownloadingFileNotFoundError("タスク履歴全件ファイルが存在しません。") from e
            raise e



import argparse
import logging
from pathlib import Path
from typing import List

import pandas

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, build_annofabapi_resource_and_login
from annofabcli.common.utils import read_multiheader_csv
from annofabcli.statistics.csv import Csv
from annofabcli.statistics.table import Table

logger = logging.getLogger(__name__)


def create_df_merged_peformance_per_user(csv_path_list: List[Path]) -> pandas.DataFrame:
    """
    `メンバごとの生産性と品質.csv` をマージしたDataFrameを返す。

    Args:
        csv_path_list:

    Returns:

    """
    df_list: List[pandas.DataFrame] = []
    for csv_path in csv_path_list:
        if csv_path.exists():
            df = read_multiheader_csv(str(csv_path))
            df_list.append(df)
        else:
            logger.warning(f"{csv_path} は存在しませんでした。")
            continue

    sum_df = df_list[0]
    for df in df_list[1:]:
        sum_df = Table.merge_productivity_per_user_from_aw_time(sum_df, df)
    return sum_df


def merge_peformance_per_user(csv_path_list: List[Path], output_path: Path):
    sum_df = create_df_merged_peformance_per_user(csv_path_list)
    csv_obj = Csv(outdir=str(output_path.parent))
    csv_obj.write_productivity_per_user(sum_df, output_path=output_path)


class MergePerfomancePerUser(AbstractCommandLineInterface):
    def main(self):
        args = self.args
        merge_peformance_per_user(csv_path_list=args.csv, output_path=args.output)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    MergePerfomancePerUser(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--csv",
        type=Path,
        nargs="+",
        required=True,
        help=("CSVファイルのパスを複数指定してください。" "CSVは、'statistics visualize'コマンドの出力結果である'メンバごとの生産性と品質.csv'と同じフォーマットです。"),
    )

    parser.add_argument("-o", "--output", type=Path, required=True, help="出力先のファイルパスを指定します。")

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "merge_peformance_per_user"
    subcommand_help = "複数の'メンバごとの生産性と品質.csv'の値をまとめて出力します。"
    description = "複数の'メンバごとの生産性と品質.csv'の値をまとめて出力します。"
    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



"""
アノテーション情報を画像化するメソッドです。
外部に公開しています。
"""
import logging
import zipfile
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional

import PIL
import PIL.Image
import PIL.ImageDraw
from annofabapi.dataclass.annotation import SimpleAnnotationDetail
from annofabapi.exceptions import AnnotationOuterFileNotFoundError
from annofabapi.models import InputData
from annofabapi.parser import SimpleAnnotationParser, lazy_parse_simple_annotation_dir, lazy_parse_simple_annotation_zip

from annofabcli.common.typing import RGB, InputDataSize

logger = logging.getLogger(__name__)

IsParserFunc = Callable[[SimpleAnnotationParser], bool]
"""アノテーションparserに対してboolを返す関数"""


def get_data_uri_of_outer_file(annotation: SimpleAnnotationDetail) -> Optional[str]:
    """
    外部ファイルの data_uri を取得する
    Args:
        annotation: アノテーション

    Returns:
        外部ファイルの data_uri

    """
    data = annotation.data
    if data is None:
        return None

    return data.get("data_uri")


def fill_annotation(
    draw: PIL.ImageDraw.Draw,
    annotation: SimpleAnnotationDetail,
    label_color_dict: Dict[str, RGB],
    outer_image: Optional[Any] = None,
) -> PIL.ImageDraw.Draw:
    """
    1個のアノテーションを、塗りつぶしで描画する。（矩形、ポリゴン、塗りつぶし、塗りつぶしv2）

    Args:
        draw: draw: (IN/OUT) PillowのDrawing Object. 変更される。
        annotation: 描画対象のアノテーション
        label_color_dict: label_nameとRGBを対応付けたdict
        outer_image: 外部ファイル（塗りつぶし）の画像オブジェクト

    Returns:
        アノテーションを描画した状態のDrawing Object
    """

    data = annotation.data
    if data is None:
        # 画像全体アノテーション
        return draw

    color = label_color_dict.get(annotation.label)
    if color is None:
        logger.warning(f"label_name = {annotation.label} のcolorが指定されていません。")
        color = (255, 255, 255)

    data_type = data["_type"]
    if data_type == "BoundingBox":
        xy = [(data["left_top"]["x"], data["left_top"]["y"]), (data["right_bottom"]["x"], data["right_bottom"]["y"])]
        draw.rectangle(xy, fill=color)

    elif data_type == "Points":
        # Polygon
        xy = [(e["x"], e["y"]) for e in data["points"]]
        draw.polygon(xy, fill=color)

    elif data_type in ["SegmentationV2", "Segmentation"]:
        # 塗りつぶしv2 or 塗りつぶし
        if outer_image is not None:
            draw.bitmap([0, 0], outer_image, fill=color)
        else:
            logger.warning(
                f"アノテーション種類が`{data_type}`ですが、`outer_image`がNoneです。 " f"annotation_id={annotation.annotation_id}"
            )

    return draw


def fill_annotation_list(
    draw: PIL.ImageDraw.Draw,
    parser: SimpleAnnotationParser,
    label_color_dict: Dict[str, RGB],
    label_name_list: Optional[List[str]] = None,
) -> PIL.ImageDraw.Draw:
    """
    1個の入力データに属するアノテーションlistを描画する

    Args:
        draw: draw: (IN/OUT) PillowのDrawing Object. 変更される。
        parser: Simple Annotationのparser
        label_color_dict: label_nameとRGBを対応付けたdict
        label_name_list: 画像化対象のlabel_name. Noneの場合は、すべてのlabel_nameが画像化対象です。

    Returns:
        アノテーションを描画した状態のDrawing Object
    """
    simple_annotation = parser.parse()

    # 下層レイヤにあるアノテーションから順に画像化する
    # reversed関数を使う理由：`simple_annotation.details`は上層レイヤのアノテーションから順に格納されているため
    if label_name_list is not None:
        annotation_list = [e for e in reversed(simple_annotation.details) if e.label in label_name_list]
    else:
        annotation_list = list(reversed(simple_annotation.details))

    for annotation in annotation_list:
        # 外部ファイルのImage情報を取得する
        data_uri_outer_image = get_data_uri_of_outer_file(annotation)
        if data_uri_outer_image is not None:
            try:
                with parser.open_outer_file(data_uri_outer_image) as f:
                    outer_image = PIL.Image.open(f)
                    # アノテーション情報を描画する
                    fill_annotation(
                        draw=draw, annotation=annotation, label_color_dict=label_color_dict, outer_image=outer_image
                    )

            except AnnotationOuterFileNotFoundError as e:
                logger.warning(str(e))
                fill_annotation(draw=draw, annotation=annotation, label_color_dict=label_color_dict, outer_image=None)

        else:
            fill_annotation(draw=draw, annotation=annotation, label_color_dict=label_color_dict, outer_image=None)

    return draw


def write_annotation_image(
    parser: SimpleAnnotationParser,
    image_size: InputDataSize,
    label_color_dict: Dict[str, RGB],
    output_image_file: Path,
    background_color: Optional[Any] = None,
    label_name_list: Optional[List[str]] = None,
):
    """
    JSONファイルに記載されているアノテーション情報を、画像化する。
    JSONファイルは、AnnoFabからダウンロードしたアノテーションzipに含まれるファイルを想定している。

    Args:
        parser: parser: Simple Annotationのparser
        image_size: 画像のサイズ. Tuple[width, height]
        label_color_dict: label_nameとRGBを対応付けたdict
        output_image_file: 出力先の画像ファイルのパス
        background_color: アノテーション画像の背景色.
            (ex) "rgb(173, 216, 230)", "#add8e6", "lightgray", (173,216,230)
            フォーマットは`ImageColor Module <https://hhsprings.bitbucket.io/docs/programming/examples/python/PIL/ImageColor.html>`_  # noqa: E501
            '指定しない場合は、黒（rgb(0,0,0)）になります。'))
        label_name_list: 画像化対象のlabel_name. Noneの場合は、すべてのlabel_nameが画像化対象です。

    Examples:
        "simple-annotation.json" を読み込み、"out.png"画像を生成する。


            from pathlib import Path
            from annofabapi.parser import SimpleAnnotationDirParser
            parser = SimpleAnnotationDirParser( "simple-annotation.json")
            label_color_dict = {"dog": (255,0,0), "bird": (0,255,0)}
            write_annotation_image(parser=parser, image_size=(64,64), label_color_dict=label_color_dict,
                               output_image_file=Path("out.png"), background_color=(64,64,64))

    """

    image = PIL.Image.new(mode="RGB", size=image_size, color=background_color)
    draw = PIL.ImageDraw.Draw(image)

    fill_annotation_list(draw, parser, label_color_dict, label_name_list=label_name_list)

    output_image_file.parent.mkdir(parents=True, exist_ok=True)
    image.save(output_image_file)


def write_annotation_images_from_path(
    annotation_path: Path,
    label_color_dict: Dict[str, RGB],
    output_dir_path: Path,
    image_size: Optional[InputDataSize] = None,
    input_data_dict: Optional[Dict[str, InputData]] = None,
    metadata_key_of_image_width: Optional[str] = None,
    metadata_key_of_image_height: Optional[str] = None,
    output_image_extension: str = "png",
    background_color: Optional[Any] = None,
    label_name_list: Optional[List[str]] = None,
    is_target_parser_func: Optional[IsParserFunc] = None,
) -> bool:
    """
    AnnoFabからダウンロードしたアノテーションzipファイル、またはそのzipを展開したディレクトリから、アノテーション情報を画像化します。
    `image_size` or `input_data_dict` のいずれかを指定する必要があります。

    Args:
        annotation_path: AnnoFabからダウンロードしたアノテーションzipファイル、またはそのzipを展開したディレクトリのパス
        label_color_dict: label_nameとRGBを対応付けたdict. Key:label_name, Value: Tuple(R,G,B)
        output_dir_path: 画像の出力先のディレクトリのパス
        image_size: 画像のサイズ. Tuple[width, height]
        input_data_dict: 入力データのDict(key:input_data_id, value:InputData)
        metadata_key_of_image_width: 入力データのメタデータのキー（画像width)
        metadata_key_of_image_height: 入力データのメタデータのキー（画像height)
        output_image_extension: 出力する画像ファイルの拡張子. 指定しない場合は"png"です。
        background_color: アノテーション画像の背景色.
            (ex) "rgb(173, 216, 230)", "#add8e6", "lightgray", (173,216,230)
            フォーマットは`ImageColor Module <https://hhsprings.bitbucket.io/docs/programming/examples/python/PIL/ImageColor.html>`_  # noqa: E501
            '指定しない場合は、黒（rgb(0,0,0)）になります。'))
        label_name_list: 画像化対象のlabel_name. Noneの場合は、すべてのlabel_nameが画像化対象です。
        is_target_parser_func: 画像化する条件を関数で指定できます。Noneの場合は、すべてを画像化します。関数の引数は`SimpleAnnotationParser`で、戻り値がbooleanです。戻り値がTrueならば画像化対象です。

    Returns:
        True: アノテーション情報の画像化に成功した。False: アノテーション情報の画像化に失敗した。

    """

    def _get_image_size(input_data_id: str) -> Optional[InputDataSize]:
        if image_size is not None:
            return image_size
        elif input_data_dict is not None:
            if metadata_key_of_image_width is not None and metadata_key_of_image_height is not None:
                input_data = input_data_dict[input_data_id]
                if input_data is None:
                    logger.warning(f"input_data_id={input_data_id}の入力データは存在しなかったので、スキップします。")
                    return None
                metadata = input_data["metadata"]
                if metadata is not None and (
                    metadata_key_of_image_width in metadata and metadata_key_of_image_height in metadata
                ):
                    try:
                        return (int(metadata[metadata_key_of_image_width]), int(metadata[metadata_key_of_image_height]))
                    except ValueError as e:
                        logger.warning(f"input_data_id={input_data_id}の入力データのメタデータの数値変換に失敗しました。{e}")
                        return None

                else:
                    logger.warning(
                        f"input_data_id={input_data_id}の入力データのメタデータに、"
                        f"{metadata_key_of_image_width}, {metadata_key_of_image_height} というキーが存在しなかったので、"
                        f"スキップします。"
                    )
                    return None
            else:
                raise ValueError(
                    f"引数`input_data_dict`がNoneでない場合は、"
                    f"`metadata_key_of_image_width`と`metadata_key_of_image_height`はnot Noneである必要があります。"
                )
        else:
            raise ValueError(f"引数`image_size`または`input_data_dict`のどちらかはnot Noneである必要があります。")

    if not annotation_path.exists():
        logger.warning(f"annotation_path: '{annotation_path}' は存在しません。")
        return False

    if annotation_path.is_dir():
        iter_lazy_parser = lazy_parse_simple_annotation_dir(annotation_path)
    elif zipfile.is_zipfile(str(annotation_path)):
        iter_lazy_parser = lazy_parse_simple_annotation_zip(annotation_path)
    else:
        logger.warning(f"annotation_path: '{annotation_path}' は、zipファイルまたはディレクトリではありませんでした。")
        return False

    count_created_image = 0
    for parser in iter_lazy_parser:
        logger.debug(f"{parser.json_file_path} を読み込みます。")
        if is_target_parser_func is not None and not is_target_parser_func(parser):
            logger.debug(f"{parser.json_file_path} の画像化をスキップします。")
            continue

        output_image_file = output_dir_path / f"{Path(parser.json_file_path).stem}.{output_image_extension}"
        tmp_image_size = _get_image_size(parser.input_data_id)
        if tmp_image_size is None:
            logger.warning(f"task_id={parser.task_id}, input_data_id={parser.input_data_id}: 画像サイズを取得できなかったので、スキップします。")
            continue

        write_annotation_image(
            parser,
            image_size=tmp_image_size,
            label_color_dict=label_color_dict,
            background_color=background_color,
            output_image_file=output_image_file,
            label_name_list=label_name_list,
        )
        logger.debug(f"画像ファイル '{str(output_image_file)}' を生成しました。")
        count_created_image += 1

    logger.info(f"{str(output_dir_path)} に、{count_created_image} 件の画像ファイルを生成しました。")
    return True



import argparse
import json
import logging
from pathlib import Path
from typing import Dict, List, Optional

import annofabapi
from annofabapi.models import TaskHistory

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, ArgumentParser, build_annofabapi_resource_and_login
from annofabcli.common.download import DownloadingFile
from annofabcli.common.enums import FormatArgument
from annofabcli.common.visualize import AddProps

logger = logging.getLogger(__name__)

TaskHistoryDict = Dict[str, List[TaskHistory]]
"""全タスクのタスク履歴一覧の集合体。keyはtask_id"""


class ListTaskHistoryWithJsonMain:
    def __init__(self, service: annofabapi.Resource):
        self.service = service
        self.facade = AnnofabApiFacade(service)

    @staticmethod
    def filter_task_history_dict(
        task_history_dict: TaskHistoryDict, task_id_list: Optional[List[str]] = None
    ) -> TaskHistoryDict:
        if task_id_list is None:
            return task_history_dict

        filtered_task_history_dict: TaskHistoryDict = {}
        for task_id in task_id_list:
            task_history_list = task_history_dict.get(task_id)
            if task_history_list is None:
                logger.warning(f"task_id='{task_id}'のタスク履歴は見つかりませんでした。")
            else:
                filtered_task_history_dict[task_id] = task_history_list
        return filtered_task_history_dict

    def get_task_history_dict(
        self, project_id: str, task_history_json: Optional[Path] = None, task_id_list: Optional[List[str]] = None
    ) -> TaskHistoryDict:
        """出力対象のタスク履歴情報を取得する"""
        if task_history_json is None:
            downloading_obj = DownloadingFile(self.service)
            cache_dir = annofabcli.utils.get_cache_dir()
            json_path = cache_dir / f"{project_id}-task_history.json"

            downloading_obj.download_task_history_json(project_id, str(json_path))
        else:
            json_path = task_history_json

        with json_path.open() as f:
            all_task_history_dict = json.load(f)

        task_history_dict = self.filter_task_history_dict(all_task_history_dict, task_id_list)

        visualize = AddProps(self.service, project_id)

        for task_history_list in task_history_dict.values():
            for task_history in task_history_list:
                visualize.add_properties_to_task_history(task_history)

        return task_history_dict

    @staticmethod
    def to_all_task_history_list_from_dict(task_history_dict: TaskHistoryDict) -> List[TaskHistory]:
        all_task_history_list = []
        for task_history_list in task_history_dict.values():
            all_task_history_list.extend(task_history_list)
        return all_task_history_list


class ListTaskHistoryWithJson(AbstractCommandLineInterface):
    def print_task_history_list(
        self,
        project_id: str,
        task_history_json: Optional[Path],
        task_id_list: Optional[List[str]],
        arg_format: FormatArgument,
    ):
        """
        タスク一覧を出力する

        Args:
            project_id: 対象のproject_id
            task_id_list: 対象のタスクのtask_id
            task_query: タスク検索クエリ
            task_list_from_json: JSONファイルから取得したタスク一覧

        """

        super().validate_project(project_id, project_member_roles=None)

        main_obj = ListTaskHistoryWithJsonMain(self.service)
        task_history_dict = main_obj.get_task_history_dict(
            project_id, task_history_json=task_history_json, task_id_list=task_id_list
        )
        if arg_format == FormatArgument.CSV:
            all_task_history_list = main_obj.to_all_task_history_list_from_dict(task_history_dict)
            self.print_according_to_format(all_task_history_list)
        else:
            self.print_according_to_format(task_history_dict)

    def main(self):
        args = self.args

        task_id_list = annofabcli.common.cli.get_list_from_args(args.task_id) if args.task_id is not None else None

        self.print_task_history_list(
            args.project_id,
            task_history_json=args.task_history_json,
            task_id_list=task_id_list,
            arg_format=FormatArgument(args.format),
        )


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ListTaskHistoryWithJson(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()
    parser.add_argument(
        "-t",
        "--task_id",
        type=str,
        nargs="+",
        help="対象のタスクのtask_idを指定します。" "`file://`を先頭に付けると、task_idの一覧が記載されたファイルを指定できます。",
    )

    parser.add_argument(
        "--task_history_json",
        type=Path,
        help="タスク履歴情報が記載されたJSONファイルのパスを指定すると、JSONに記載された情報を元にタスク履歴一覧を出力します。"
        "指定しない場合、JSONファイルをダウンロードします。"
        "JSONファイルは`$ annofabcli project download task_history`コマンドで取得できます。",
    )

    argument_parser.add_format(
        choices=[FormatArgument.CSV, FormatArgument.JSON, FormatArgument.PRETTY_JSON],
        default=FormatArgument.CSV,
    )
    argument_parser.add_output()
    argument_parser.add_csv_format()

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "list_with_json"
    subcommand_help = "タスク履歴の一覧を出力します。"
    description = "タスク履歴の一覧を出力します。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



import argparse
import logging
from typing import Any, Dict, List

import annofabapi
import pandas

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, ArgumentParser, build_annofabapi_resource_and_login
from annofabcli.common.enums import FormatArgument

logger = logging.getLogger(__name__)


class ListTaskCreationHistoryMain:
    def __init__(self, service: annofabapi.Resource):
        self.service = service
        self.facade = AnnofabApiFacade(service)

    def get_data_list(self, project_id: str) -> List[Dict[str, Any]]:
        def create_elm(job: Dict[str, Any]) -> Dict[str, Any]:
            job_detail = job["job_detail"]
            return {
                "project_id": job["project_id"],
                "job_id": job["job_id"],
                "job_status": job["job_status"],
                "generated_task_count": job_detail["generated_task_count"],
                "created_datetime": job["created_datetime"],
                "updated_datetime": job["updated_datetime"],
                "task_generated_rule": job_detail["request"]["task_generate_rule"],
            }

        query_params = {"type": "gen-tasks"}
        job_list = self.service.wrapper.get_all_project_job(project_id, query_params=query_params)
        return [create_elm(job) for job in job_list]


class ListTaskCreationHistory(AbstractCommandLineInterface):
    def main(self):
        args = self.args
        main_obj = ListTaskCreationHistoryMain(self.service)
        data_list = main_obj.get_data_list(args.project_id)
        df = pandas.DataFrame(data_list)
        self.print_according_to_format(df)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ListTaskCreationHistory(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()

    argument_parser.add_format(
        choices=[FormatArgument.CSV, FormatArgument.JSON, FormatArgument.PRETTY_JSON], default=FormatArgument.CSV
    )
    argument_parser.add_output()
    argument_parser.add_csv_format()

    argument_parser.add_query()
    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "list_task_creation_history"
    subcommand_help = "タスクの作成履歴一覧を出力します。"
    description = "タスクの作成履歴一覧を出力します。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



import argparse

import annofabcli
import annofabcli.annotation_specs.list_annotation_specs_history
import annofabcli.annotation_specs.list_annotation_specs_label
import annofabcli.annotation_specs.print_label_color
import annofabcli.common.cli


def parse_args(parser: argparse.ArgumentParser):

    subparsers = parser.add_subparsers(dest="subcommand_name")

    # サブコマンドの定義
    annofabcli.annotation_specs.list_annotation_specs_history.add_parser(subparsers)
    annofabcli.annotation_specs.list_annotation_specs_label.add_parser(subparsers)
    annofabcli.annotation_specs.print_label_color.add_parser(subparsers)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "annotation_specs"
    subcommand_help = "アノテーション仕様関係のサブコマンド"
    description = "アノテーション仕様関係のサブコマンド"

    parser = annofabcli.common.cli.add_parser(
        subparsers, subcommand_name, subcommand_help, description, is_subcommand=False
    )
    parse_args(parser)



import argparse
import datetime
import json
import logging
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple

import annofabapi
from annofabapi.dataclass.input import InputData

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, ArgumentParser, build_annofabapi_resource_and_login
from annofabcli.common.dataclasses import WaitOptions
from annofabcli.common.download import DownloadingFile
from annofabcli.common.enums import FormatArgument
from annofabcli.common.facade import InputDataQuery, match_input_data_with_query

logger = logging.getLogger(__name__)

DatetimeRange = Tuple[Optional[datetime.datetime], Optional[datetime.datetime]]


class ListInputDataWithJsonMain:
    def __init__(self, service: annofabapi.Resource):
        self.service = service
        self.facade = AnnofabApiFacade(service)

    @staticmethod
    def filter_input_data_list(
        input_data: Dict[str, Any],
        input_data_id_set: Optional[Set[str]] = None,
        input_data_query: Optional[InputDataQuery] = None,
    ) -> bool:
        result = True

        dc_input_data = InputData.from_dict(input_data)
        result = result and match_input_data_with_query(dc_input_data, input_data_query)
        if input_data_id_set is not None:
            result = result and (dc_input_data.input_data_id in input_data_id_set)
        return result

    def get_input_data_list(
        self,
        project_id: str,
        input_data_json: Optional[Path],
        input_data_id_list: Optional[List[str]] = None,
        input_data_query: Optional[InputDataQuery] = None,
        is_latest: bool = False,
        wait_options: Optional[WaitOptions] = None,
    ) -> List[Dict[str, Any]]:
        if input_data_json is None:
            downloading_obj = DownloadingFile(self.service)
            cache_dir = annofabcli.utils.get_cache_dir()
            json_path = cache_dir / f"{project_id}-input_data.json"

            downloading_obj.download_input_data_json(
                project_id, str(json_path), is_latest=is_latest, wait_options=wait_options
            )
        else:
            json_path = input_data_json

        # TODO logger json load
        with json_path.open() as f:
            input_data_list = json.load(f)

        input_data_id_set = set(input_data_id_list) if input_data_id_list is not None else None
        filtered_input_data_list = [
            e
            for e in input_data_list
            if self.filter_input_data_list(e, input_data_query=input_data_query, input_data_id_set=input_data_id_set)
        ]
        return filtered_input_data_list


class ListInputDataWithJson(AbstractCommandLineInterface):
    def main(self):
        args = self.args

        input_data_id_list = (
            annofabcli.common.cli.get_list_from_args(args.input_data_id) if args.input_data_id is not None else None
        )
        input_data_query = (
            InputDataQuery.from_dict(annofabcli.common.cli.get_json_from_args(args.input_data_query))
            if args.input_data_query is not None
            else None
        )
        wait_options = (
            WaitOptions.from_dict(annofabcli.common.cli.get_json_from_args(args.wait_options))
            if args.wait_options is not None
            else None
        )

        project_id = args.project_id
        super().validate_project(project_id, project_member_roles=None)

        main_obj = ListInputDataWithJsonMain(self.service)
        input_data_list = main_obj.get_input_data_list(
            project_id=project_id,
            input_data_json=args.input_data_json,
            input_data_id_list=input_data_id_list,
            input_data_query=input_data_query,
            is_latest=args.latest,
            wait_options=wait_options,
        )

        logger.debug(f"入力データ一覧の件数: {len(input_data_list)}")

        if len(input_data_list) > 0:
            self.print_according_to_format(input_data_list)
        else:
            logger.info(f"入力データ一覧の件数が0件のため、出力しません。")


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ListInputDataWithJson(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()

    parser.add_argument(
        "-iq",
        "--input_data_query",
        type=str,
        help="入力データの検索クエリをJSON形式で指定します。"
        "`file://`を先頭に付けると、JSON形式のファイルを指定できます。"
        "指定できるキーは、`input_data_name`, `input_data_path`です。",
    )

    parser.add_argument(
        "-i",
        "--input_data_id",
        type=str,
        nargs="+",
        help="対象のinput_data_idを指定します。" "`file://`を先頭に付けると、input_data_idの一覧が記載されたファイルを指定できます。",
    )

    parser.add_argument(
        "--input_data_json",
        type=Path,
        help="入力データ情報が記載されたJSONファイルのパスを指定すると、JSONに記載された情報を元に入力データ一覧を出力します。"
        "指定しない場合、全件ファイルをダウンロードします。"
        "JSONファイルは`$ annofabcli project download input_data`コマンドで取得できます。",
    )

    parser.add_argument(
        "--latest", action="store_true", help="最新の入力データ一覧ファイルを参照します。このオプションを指定すると、入力データ一覧ファイルを更新するのに約5分以上待ちます。"
    )

    parser.add_argument(
        "--wait_options",
        type=str,
        help="入力データ一覧ファイルの更新が完了するまで待つ際のオプションを、JSON形式で指定してください。"
        "`file://`を先頭に付けるとjsonファイルを指定できます。"
        'デフォルは`{"interval":60, "max_tries":360}` です。'
        "`interval`:完了したかを問い合わせる間隔[秒], "
        "`max_tires`:完了したかの問い合わせを最大何回行うか。",
    )

    argument_parser.add_format(
        choices=[
            FormatArgument.CSV,
            FormatArgument.JSON,
            FormatArgument.PRETTY_JSON,
            FormatArgument.INPUT_DATA_ID_LIST,
        ],
        default=FormatArgument.CSV,
    )
    argument_parser.add_output()
    argument_parser.add_csv_format()

    argument_parser.add_query()
    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "list_with_json"
    subcommand_help = "入力データ全件ファイルから一覧を出力します。"
    description = "入力データ全件ファイルから一覧を出力します。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



"""
プロジェクトから複数のユーザを削除する。
"""
import argparse
import logging
from typing import List

import requests

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, build_annofabapi_resource_and_login

logger = logging.getLogger(__name__)


class DeleteUser(AbstractCommandLineInterface):
    """
    ユーザをプロジェクトから削除する
    """

    def drop_role_with_organization(self, organization_name: str, user_id_list: List[str]):
        projects = self.service.wrapper.get_all_projects_of_organization(
            organization_name, query_params={"account_id": self.service.api.account_id}
        )

        for project in projects:
            project_id = project["project_id"]
            project_title = project["title"]

            try:
                if not self.facade.my_role_is_owner(project_id):
                    logger.warning(
                        f"オーナではないため、プロジェクトメンバを脱退させられません。" f"project_id = {project_id}, project_tilte = {project_title}"
                    )
                    continue

                self.service.wrapper.drop_role_to_project_members(project_id, user_id_list)
                logger.info(f"{project_title} からユーザを脱退させました。project_id = {project_id}")

            except requests.exceptions.HTTPError as e:
                logger.warning(e)
                logger.warning(f"エラーのため、{project_title} からユーザを脱退させられませんでした。")

    def drop_role_with_project_id(self, project_id_list: List[str], user_id_list: List[str]):
        for project_id in project_id_list:

            try:
                if not self.facade.my_role_is_owner(project_id):
                    logger.warning(f"オーナではないため、プロジェクトメンバを脱退させられません。" f"project_id = {project_id}")
                    continue

                project_title = self.service.api.get_project(project_id)[0]["title"]
                self.service.wrapper.drop_role_to_project_members(project_id, user_id_list)
                logger.info(f"{project_title} からユーザを脱退させました。project_id={project_id}")

            except requests.exceptions.HTTPError as e:
                logger.warning(e)
                logger.warning(f"エラーのため、project_id = {project_id} からユーザを脱退させられませんでした。")

    def main(self):
        args = self.args
        user_id_list = annofabcli.common.cli.get_list_from_args(args.user_id)

        if args.organization is not None:
            self.drop_role_with_organization(args.organization, user_id_list)

        elif args.project_id is not None:
            project_id_list = annofabcli.common.cli.get_list_from_args(args.project_id)
            self.drop_role_with_project_id(project_id_list, user_id_list)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    DeleteUser(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    parser.add_argument(
        "-u",
        "--user_id",
        type=str,
        nargs="+",
        required=True,
        help="脱退させるユーザのuser_idを指定してください。`file://`を先頭に付けると、一覧が記載されたファイルを指定できます。",
    )

    drop_group = parser.add_mutually_exclusive_group(required=True)
    drop_group.add_argument(
        "-p",
        "--project_id",
        type=str,
        nargs="+",
        help="脱退させるプロジェクトのproject_idを指定してください。`file://`を先頭に付けると、一覧が記載されたファイルを指定できます。",
    )

    drop_group.add_argument(
        "-org", "--organization", type=str, help="組織名を指定すると、組織配下のすべてのプロジェクト（自分が所属している）から、ユーザを脱退させます。"
    )

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "delete"
    subcommand_help = "複数のプロジェクトから、ユーザを脱退させます。"
    description = "複数のプロジェクトから、ユーザを脱退させます。"
    epilog = "オーナロールを持つユーザで実行してください。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)
